import tempfile
import unittest
import os
import sys
import numpy as np
import basico
import shutil

class TestBasicoParamterEstimation(unittest.TestCase):

    def setUp(self):
        self.model = basico.load_example('LM')
        self.assertTrue(self.model.getModel().getObjectName() ==
                        'Kinetics of a  Michaelian enzyme measured spectrophotometrically')

    def tearDown(self):
        basico.remove_datamodel(self.model)

    def test_get_parameters(self):
        params = basico.get_parameters()
        signal = params.loc['signal']
        expression = signal.expression
        self.assertTrue(expression != '')
        conversion = basico.model_info._replace_names_with_cns(expression)
        self.assertTrue(conversion != '')
        expression = basico.model_info._replace_cns_with_names(conversion)
        self.assertTrue(expression != '')

    def test_get_data(self):
        data = basico.get_experiment_data_from_model()
        self.assertEqual(len(data), 5)
        set_1 = data[0]
        self.assertEqual(set_1.shape, (100, 3))

        names = basico.get_experiment_names()
        self.assertIn('Experiment', names)

        files = basico.get_experiment_filenames()
        self.assertEqual(len(files), len(names))

        set_2 = basico.get_data_from_experiment('Experiment')
        self.assertEqual(set_2.shape, (100, 3))

        sol_before = basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION)

        # get statstic
        stat = basico.get_fit_statistic(include_parameters=True, include_fitted=True)

        # remove data
        basico.remove_experiments()

        files_to_delete = []
        # add data back
        for exp, name in zip(data, names):
            files_to_delete.append(basico.add_experiment(name, exp))

        # remove data
        basico.remove_experiments()

        # since we have affected experiments, reset them
        basico.set_fit_parameters(basico.get_fit_parameters())

        # save to temp file
        main_file = tempfile.mktemp()
        basico.save_model_and_data(main_file, delete_data_on_exit=True)

        # ensure it still works
        sol_after = basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION)
        self.assertListEqual(basico.as_dict(sol_before[['sol']]), basico.as_dict(sol_after[['sol']]))

        # delete temp files
        for file in files_to_delete:
            os.remove(file)
        os.remove(main_file)


    def test_mapping(self):
        mapping = basico.get_experiment_mapping(0)
        experiments = basico.save_experiments_to_dict()
        self.assertIsNotNone(experiments)
        # remove cns
        for entry in experiments[0]['mapping']:
            if 'cn' in entry:
                del entry['cn']

        basico.load_experiments_from_dict(experiments)
        experiments2 = basico.save_experiments_to_dict()
        self.assertIsNotNone(experiments)
        self.assertEqual(experiments[0]['mapping'][0]['object'], experiments2[0]['mapping'][0]['object'])
        self.assertEqual(experiments[0]['mapping'][2]['object'], experiments2[0]['mapping'][2]['object'])


    def test_experiment_dict(self):
        exp1 = basico.get_experiment_dict(0)
        self.assertTrue(exp1 is not None)
        # save to yaml
        yaml_str = basico.save_experiments_to_yaml()
        self.assertTrue(yaml_str is not None)
        # restore from yaml
        basico.load_experiments_from_yaml(yaml_str)
        # save again
        yaml_str2 = basico.save_experiments_to_yaml()
        self.assertEqual(yaml_str, yaml_str2)

    @unittest.skipIf(sys.version_info < (3, 6, 0), 'This test requires assertLogs which is not available before')
    def test_change_bounds(self):
        basico.set_fit_parameters([{'name': '(R1).k2', 'lower': 1, 'upper': 0.01}])
        with self.assertLogs(level='ERROR') as _:
            sol = basico.as_dict(basico.run_parameter_estimation())
        self.assertTrue(np.isnan(sol['sol']))

    def test_get_plotting_data(self):
        exp, sim = basico.get_simulation_results()
        self.assertTrue(len(exp) == len(sim))
        exp, sim = basico.get_simulation_results(values_only=True)
        self.assertListEqual(exp[0].Time.to_list(), sim[0].reset_index().Time.to_list())

    def test_get_item_template(self):
        template = basico.get_fit_item_template(include_local=True, include_global=True)
        self.assertTrue(len(template) == 5)

    def test_get_settings(self):
        settings = basico.get_task_settings('Parameter Estimation')
        self.assertTrue('method' in settings)
        self.assertTrue('Iteration Limit' in settings['method'])
        self.assertEqual(settings['method']['Iteration Limit'], 500)
        basico.set_task_settings('Parameter Estimation', {'method': {'Iteration Limit': 600}})
        settings = basico.get_task_settings('Parameter Estimation')
        self.assertTrue('method' in settings)
        self.assertTrue('Iteration Limit' in settings['method'])
        self.assertEqual(settings['method']['Iteration Limit'], 600)

    def test_run(self):
        sol = basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION)
        self.assertTrue(sol is not None)
        self.assertAlmostEqual(sol.loc['(R1).k2'].sol, 1.0, places=2)

    def test_remove(self):
        fit_items = basico.get_fit_parameters()
        self.assertIsNotNone(fit_items)
        basico.remove_fit_parameters()
        fit_items_2 = basico.get_fit_parameters()
        self.assertIsNone(fit_items_2)
        basico.set_fit_parameters(fit_items)

    def test_save(self):
        cps_name = tempfile.mktemp()
        basico.save_model_and_data(cps_name, delete_data_on_exit=True)
        d2 = basico.load_model(cps_name)
        self.assertIsNotNone(d2)
        basico.remove_datamodel(d2)


class TestBasicoParamterEstimationPK(unittest.TestCase):

    def setUp(self):
        self.model = basico.load_example('PK')
        self.assertTrue(self.model.getModel().getObjectName() ==
                        'Pritchard2002_glycolysis')

    def tearDown(self):
        basico.remove_datamodel(self.model)

    def test_get_data(self):

        # without renaming data should remain as is
        experiment = basico.get_data_from_experiment(0, rename_headers=False)
        self.assertEqual(['# Time', 'Values[F16BP_obs]', 'Values[Glu_obs]', 'Values[Pyr_obs]',
                         '[Glc(ext)]_0', 'Unnamed: 5'], list(experiment.columns))
        self.assertEqual((101, 6), experiment.shape)

        # data should be renamed as expected, and unassigned columns dropped
        experiment = basico.get_data_from_experiment(0, rename_headers=True)
        self.assertEqual(['Time', '[Fru1,6-P2]', '[Glc(int)]', '[pyruvate]', '[Glc(ext)]_0'],
                         list(experiment.columns))
        self.assertEqual((101, 5), experiment.shape)
        self.assertEqual(0, experiment.iloc[0][0])
        self.assertEqual(2, experiment.iloc[100][0])

        # now simulate
        exp, sim = basico.get_simulation_results()
        self.assertTrue(len(exp) == len(sim) == 2)

        # ensure that simulations are different
        self.assertGreater(float(((sim[0][['Fru1,6-P2']] - sim[1][['Fru1,6-P2']]) ** 2).sum()), 10)

        # plotting is tested manually only
        # basico.plot_per_experiment()
        # basico.plot_per_dependent_variable()
        # matplotlib.pyplot.show()

    def test_set_fit_items(self):
        old_fit_items = basico.get_fit_parameters().reset_index()
        basico.set_fit_parameters(old_fit_items)
        fit_items = basico.get_fit_parameters().reset_index()
        old_first = old_fit_items.iloc[0].to_dict()
        new_first = fit_items.iloc[0].to_dict()
        self.assertEqual(old_first['name'], new_first['name'])
        self.assertEqual(old_first['lower'], new_first['lower'])
        self.assertEqual(old_first['upper'], new_first['upper'])

    def test_statistic(self):
        # when no fit is performed, the statistic should not crash
        result = basico.get_fit_statistic(include_parameters=True)
        self.assertTrue(result is not None)
        self.assertEqual(result['f_evals'], 0)

        # now get actual fit
        basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION)
        result = basico.get_fit_statistic()
        self.assertTrue(result is not None)
        self.assertTrue(result['obj'] < 0.2)

    def test_get_data(self):
        # contains all experimental data
        data = basico.get_experiment_data_from_model()
        self.assertTrue(data is not None)

        # to get data only from one experiment, you'd call
        for name in basico.get_experiment_names():
            data = basico.get_data_from_experiment(name)
            self.assertTrue(data is not None)
            # can be renamed
            data = basico.get_data_from_experiment(name, rename_headers=True)
            self.assertTrue(data is not None)

class TestMultipleResults(unittest.TestCase):

    def setUp(self):
        self.model = basico.load_model(
            os.path.join(os.path.dirname(__file__), 'multiple_experiments.cps')
        )
        self.assertTrue(self.model.getModel().getObjectName() ==
                        'multiple_experiments')

    def tearDown(self):
        basico.remove_datamodel(self.model)

    def test_current_solution(self):
        result = basico.get_simulation_results()
        self.assertTrue(len(result) == 2)
        self.assertTrue(len(result[0]) == len(result[0]))


if __name__ == '__main__':
    unittest.main()



"""A submodule for accessing the BioModels API.

This submodule accesses the BioModels REST api as described on:

    <https://www.ebi.ac.uk/biomodels/docs/>

Examples:

    >>>  # get info for a specific model
    >>>  info = get_model_info(12)
    >>>  print(info['name'], info['files']['main'][0]['name'])

    >>>  # get all files for one model
    >>>  files = get_files_for_model(12)
    >>>  print(files['main'][0]['name'])

    >>>  # get content of specific model
    >>>  sbml = get_content_for_model(12)
    >>>  print(sbml)

    >>>  # search for model
    >>>  models = search_for_model('repressilator')
    >>>  for model in models:
    >>>     print(model['id'], model['name'], model['format'])


"""

try:
    import sys
    if sys.version_info.major == 2:
        import urllib as urllib2
        from urllib import quote_plus
    else:
        import urllib2
        from urllib2 import quote_plus
    _use_urllib2 = True
except ImportError:
    import urllib
    import urllib.request
    from urllib.parse import quote_plus
    _use_urllib2 = False

import json


END_POINT = 'https://www.ebi.ac.uk/biomodels/'


def download_from(url):
    """Convenience method reading content from a URL.

    This convenience method uses urlopen on either python 2.7 or 3.x

    :param url: the url to read from
    :type url: str
    :return: the contents of the URL as str
    :rtype: str
    """
    if _use_urllib2:
        content = urllib2.urlopen(url).read()
    else:
        content = urllib.request.urlopen(url).read().decode('utf-8')
    return content


def download_json(url):
    """Convenience method reading the content of the url as JSON.

    :param url: the url to read from
    :type url: str
    :return: a python object representing the json content loaded
    :rtype: dict
    """
    content = download_from(url)
    return json.loads(content)


def get_model_info(model_id):
    """Return the model info for the provided `model_id`.

    :param model_id: either an integer, or a valid model id
    :return: a python object describing the model
    """
    if type(model_id) is int:
        model_id = 'BIOMD{0:010d}'.format(model_id)
    result = download_json(END_POINT + model_id + '?format=json')
    return result


def get_files_for_model(model_id):
    """Retrieves the json structure for all files for the given biomodel.

    The structure is of form:

    >>> get_files_for_model(10)
    {
        'additional': [
            {'description': 'Auto-generated Scilab file',
             'fileSize': '3873',
             'name': 'BIOMD0000000010.sci'},
             ...
        ],
        'main': [
            {'fileSize': '31568',
             'name': 'BIOMD0000000010_url.xml'
            }
        ]
    }

    :param model_id: the model id (as int or string)
    :return: json structure
    """
    if type(model_id) is int:
        model_id = 'BIOMD{0:010d}'.format(model_id)
    result = download_json(END_POINT + 'model/files/' + model_id + '?format=json')
    return result


def get_content_for_model(model_id, file_name=None):
    """Downloads the specified file from biomodels

    :param model_id: the model id as int, or string
    :param file_name: the filename to download (or None, to download the main file)
    :return: the content of the specified file
    """
    if type(model_id) is int:
        model_id = 'BIOMD{0:010d}'.format(model_id)
    if file_name is None:
        file_name = get_files_for_model(model_id)['main'][0]['name']
    return download_from(END_POINT + 'model/download/' + model_id + '?filename=' + file_name)


def search_for_model(query, offset=0, num_results=10, sort='id-asc'):
    """Queries the biomodel database

    Queries the database, for information about the query system see:
    <https://www.ebi.ac.uk/biomodels-static/jummp-biomodels-help/model_search.html>

    Example:

        >>> search_for_model('glycolysis')
        [...,
            {
            'format': 'SBML',
            'id': 'BIOMD0000000206',
            'lastModified': '2012-07-04T23:00:00Z',
            'name': 'Wolf2000_Glycolytic_Oscillations',
            'submissionDate': '2008-11-27T00:00:00Z',
            'submitter': 'Harish Dharuri',
            'url': 'https://www.ebi.ac.uk/biomodels/BIOMD0000000206'
            }
        ]

        Note by default, it will include only manually curated models, to obtain Non-curated models you would use:

        >>> search_for_model('Hodgkin AND curationstatus:"Non-curated"')
        [...,
            {
            'format': 'SBML',
            'id': 'MODEL1006230012',
            'lastModified': '2012-02-02T00:00:00Z'',
            'name': 'Stewart2009_ActionPotential_PurkinjeFibreCells',
            'submissionDate': '2010-06-22T23:00:00Z',
            'submitter': 'Camille Laibe',
            'url': 'https://www.ebi.ac.uk/biomodels/MODEL1006230012'
            }
        ]

    :param query: the query to use (it will be encoded with quote_plus before send out, so it is safe to use spaces)
    :param offset: offset (defaults to 0)
    :param num_results: number of results to obtain (defaults to 10)
    :param sort: sort criteria to be used (defaults to id-asc)
    :return: the search result as [{}]
    """
    url = END_POINT + 'search?query=' + quote_plus(query)
    url += '&offset=' + str(offset)
    url += "&numResults=" + str(num_results)
    url += '&sort=' + sort
    result = download_json(url + '&format=json')
    return result['models']



import sys

import unittest
import glob
import logging

import numpy as np
import pandas as pd
import os.path
import yaml

# set log level to debug
# logging.basicConfig(level=logging.DEBUG)
_REMOVE_TEMP_FILES = True


import basico

try:
    import petab_select
    import petab
    from petab import Problem
    from basico.petab import PetabSimulator, evaluate_problem
    petab_test_enabled = True
except ImportError:
    # print import error
    import traceback
    traceback.print_exc()
    petab_test_enabled = False
    pass


_dir_name = os.path.dirname(__file__)
_BENCHMARK_MODEL_DIR = os.getenv('PETAB_BENCHMARK_MODELS', None)
_PETAB_SELECT_MODEL_DIR = os.getenv('PETAB_SELECT_MODELS', None)


class PetabTestCase(unittest.TestCase):
    @unittest.skipUnless(petab_test_enabled and
                         _BENCHMARK_MODEL_DIR and
                         os.path.exists(_BENCHMARK_MODEL_DIR),
                         "require petab and benchmark models for this test")
    def test_import(self):
        files = glob.glob(_BENCHMARK_MODEL_DIR + '/*/*.yaml')
        for f in files:
            if 'Sneyd_PNAS2002' not in f:
                continue
            #if 'Lucarelli' not in f:
            #    continue
            logging.warning(f)
            pp = Problem.from_yaml(f)
            name = os.path.splitext(os.path.basename(f))[0]
            sim = PetabSimulator(pp, working_dir='./test_data/', name=name)
            sim.simulate()

    @unittest.skipUnless(petab_test_enabled, "require petab for these tests")
    def test_dataframe_convert(self):
        mes_df = pd.read_csv(os.path.join(_dir_name, 'mes_df.csv'))
        sim_df = pd.read_csv(os.path.join(_dir_name, 'sim_df.csv'))
        sim_df = sim_df.set_index('Time')

        result_df = mes_df.copy(True)
        result_df = result_df.rename(columns={"measurement": "simulation"})
        basico.petab.core._update_df_from_simulation(result_df, sim_df, 'model1_data1')
        result_df.to_csv('out_df.csv', index=False)        
        os.remove('out_df.csv')
        self.assertEqual(mes_df.shape[0], result_df.shape[0])
        self.assertEqual(mes_df.shape[1], result_df.shape[1])

    @unittest.skipUnless(petab_test_enabled, "require petab for these tests")
    def test_dataframe_transform(self):
        log_df = pd.read_csv(os.path.join(_dir_name, 'simulation_log.tsv'), sep='\t')
        nom_df = pd.read_csv(os.path.join(_dir_name, 'simulation_nom.tsv'), sep='\t')

        result_df = basico.petab.transform_simulation_df(
            log_df,
            pd.DataFrame(
                data=[{
                    'observableId': 'observable_x_0ac',
                    'observableFormula': 'observable_x_0ac',
                    'observableTransformation': 'log'
                 }]
         ))

        transformed = result_df.query('observableId == "observable_x_0ac"')
        normal = nom_df.query('observableId == "observable_x_0ac"')
        self.assertTrue(np.allclose(
            transformed['simulation'].values, normal['simulation'].values))

    @unittest.skipUnless(petab_test_enabled, "require petab for these tests")
    def test_dataframe_convert_steady(self):
        mes_df = pd.read_csv(os.path.join(_dir_name, 'ss_mes_df.csv'))
        sim_df = pd.read_csv(os.path.join(_dir_name, 'steady_df.csv'))

        result_df = mes_df.copy(True)
        result_df = result_df.rename(columns={"measurement": "simulation"})
        basico.petab.core._update_df_from_simulation(result_df, sim_df, 'control')
        result_df.to_csv('ss_out_df.csv', index=False)
        os.remove('ss_out_df.csv')
        self.assertEqual(mes_df.shape[0], result_df.shape[0])
        self.assertEqual(mes_df.shape[1], result_df.shape[1])

    @unittest.skipUnless(petab_test_enabled, "require petab for these tests")
    def test_dataframe_multiple_times(self):
        mes_df = pd.read_csv(os.path.join(_dir_name, 'ex_0008.tsv'), sep='\t')
        sim_df = pd.read_csv(os.path.join(_dir_name, '0008.tsv'), sep='\t', index_col='Time')

        result_df = mes_df.copy(True)
        result_df = result_df.rename(columns={"measurement": "simulation"})
        basico.petab.core._update_df_from_simulation(result_df, sim_df, 'c0')
        result_df.to_csv('out0008.csv', index=False)
        os.remove('out0008.csv')
        self.assertEqual(mes_df.shape[0], result_df.shape[0])
        self.assertEqual(mes_df.shape[1], result_df.shape[1])


    @unittest.skipUnless(petab_test_enabled, "require petab for these tests")
    def test_model_selection(self):
        problem = petab_select.Problem.from_yaml(os.path.join(_dir_name, 'model_selection', 'petab_select_problem.yaml'))
        best = evaluate_problem(problem, temp_dir=os.path.join(_dir_name, 'temp_selection'),
                                delete_temp_files=_REMOVE_TEMP_FILES)
        self.assertIsNotNone(best)

    @unittest.skipUnless(petab_test_enabled and
                         _PETAB_SELECT_MODEL_DIR and
                         os.path.exists(_PETAB_SELECT_MODEL_DIR),
                         "require petab for these tests")
    def test_model_selection_test9_expected(self):
        # logging.basicConfig(level=logging.DEBUG)
        expected_file = os.path.join(_PETAB_SELECT_MODEL_DIR, '0009/expected.yaml')
        self.assertTrue(os.path.exists(expected_file))

        model = petab_select.Model.from_yaml(expected_file)
        nllh_before = model.get_criterion(petab_select.Criterion.NLLH)
        model.criteria = {}
        basico.petab.evaluate_model(model, temp_dir=os.path.join(_dir_name, 'temp_selection'), delete_temp_files=_REMOVE_TEMP_FILES)
        nllh_after = model.get_criterion(petab_select.Criterion.NLLH)
        self.assertAlmostEqual(nllh_before, nllh_after, places=3)

    @unittest.skipUnless(petab_test_enabled and
                         _PETAB_SELECT_MODEL_DIR and
                         os.path.exists(_PETAB_SELECT_MODEL_DIR),
                         "require petab for these tests")
    def test_model_selection_suite(self):
        # logging.basicConfig(level=logging.DEBUG)

        files = sorted(glob.glob(_PETAB_SELECT_MODEL_DIR+'/*/peta*.yaml'))
        self.assertTrue(len(files) > 0)
        for f in files:
            if '0009' in f:
                # skip test #9, because it is taking too long
                continue

            problem = petab_select.Problem.from_yaml(f)
            best = evaluate_problem(problem,
                                    temp_dir=os.path.join(_dir_name, 'temp_selection'),
                                    delete_temp_files=_REMOVE_TEMP_FILES)
            self.assertIsNotNone(best)
            # print(os.path.dirname(f), best.model_subspace_id, best.criteria)

            # read expected file
            expected_file = os.path.join(os.path.dirname(f), 'expected.yaml')
            # parse yaml file
            with open(expected_file, 'r') as stream:
                expected = yaml.safe_load(stream)
            self.assertEqual(expected['model_subspace_id'], best.model_subspace_id)
            self.assertListEqual(expected['model_subspace_indices'], best.model_subspace_indices)
            for p in best.estimated_parameters:
                self.assertAlmostEqual(expected['estimated_parameters'][p], best.estimated_parameters[p], 5)


if __name__ == '__main__':
    unittest.main()




"""This module is concerned with brining the model to steady state.

The :func:`run_steadystate` method brings the currently loaded model to steady state
supporting a number of parameters you could set. Once done, additional calls to functions
like :func:`.get_species` or :func:`.get_reactions` will contain the steady state values.

Example:

    >>> run_steadystate()
    >>> get_species()

"""
import COPASI
from . import model_io
from . import model_info
import logging

logger = logging.getLogger(__name__)


def run_steadystate(**kwargs):
    """Brings the model to steady state.

    The function will use the :func:`.get_current_model()` unless one is provided.

    :param kwargs: optional arguments:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

     - `use_initial_values` (bool): whether to use initial values

     - `scheduled` (bool): sets whether the task is scheduled or not

     - `update_model` (bool): sets whether the model should be updated, or reset to initial conditions.

     - | `criterion` (str): specifies the acceptance criterion to be used for a steady state can be one of:
       |
       |  * `Distance and Rate`: both the Distance and the Rate criterion have to be fulfilled to accept
       |  * `Distance`: the distance criterion
       |  * `Rate`: the rate of change has to be sufficiently small


     - `settings` (dict): a dictionary with settings to use, in the same format as the ones obtained from
                         :func:`.get_task_settings`

    :return: integer status information whether the steady state was reached:

       - `0`: not found
       - `1`: steady state found
       - `2`: equillibrium steady state found
       - `3`: steady state with negative concentrations found

    :rtype: int
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    model.getModel().compileIfNecessary()

    task = model.getTask('Steady-State')
    assert (isinstance(task, COPASI.CSteadyStateTask))

    if 'scheduled' in kwargs:
        task.setScheduled(kwargs['scheduled'])

    if 'update_model' in kwargs:
        task.setUpdateModel(kwargs['update_model'])

    if 'criterion' in kwargs:
        method = task.getMethod()
        assert (isinstance(method, COPASI.CSteadyStateMethod))
        method.getParameter('Target Criterion').setStringValue(kwargs['criterion'])

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSteadyStateProblem))

    use_initial_values = kwargs.get('use_initial_values', True)

    if 'settings' in kwargs:
        model_info.set_task_settings(task, kwargs['settings'])

    result = task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI)
    if not result:
        logger.error("Error while initializing the simulation: " +
                      COPASI.CCopasiMessage.getLastMessage().getText())
    else:
        result = task.processRaw(use_initial_values)
        if not result:
            logger.error("Error while running the simulation: " +
                          COPASI.CCopasiMessage.getLastMessage().getText())

    return task.getResult()



import glob
import shutil
import tempfile
import unittest
import basico.task_profile_likelihood as pl
import basico
import pandas as pd
import os

dir_name = os.path.dirname(__file__)


class TestProfile(unittest.TestCase):
    def test_adjustments(self):
        self.assertAlmostEqual(2, pl._adjust_value(1, '200%'))
        self.assertAlmostEqual(1, pl._adjust_value(2, '-50%'))
        self.assertAlmostEqual(3, pl._adjust_value(2, '+50%'))
        self.assertAlmostEqual(0.1, pl._adjust_value(0.2, '-50%'))
        self.assertAlmostEqual(0.3, pl._adjust_value(0.2, '+50%'))
        self.assertAlmostEqual(3, pl._adjust_value(2, '150%'))
        self.assertAlmostEqual(3, pl._adjust_value(2, 1.5))
        self.assertAlmostEqual(1, pl._adjust_value(2, 0.5))
        self.assertAlmostEqual(5, pl._adjust_value(3, '=5'))
        self.assertAlmostEqual(0.3, pl._adjust_value(3, '=10%'))
        self.assertAlmostEqual(-0.3, pl._adjust_value(3, '=-10%'))
        self.assertAlmostEqual(1, pl._adjust_value(3, None, 1))
        self.assertAlmostEqual(1, pl._adjust_value(3, None, '1'))
        self.assertAlmostEqual(1, pl._adjust_value(3, 'default', '1'))
        self.assertAlmostEqual(1, pl._adjust_value(3, 'default', 1))
        self.assertAlmostEqual(1, pl._adjust_value(3, '-1SD', 1, 2))
        self.assertAlmostEqual(7, pl._adjust_value(3, '2SD', 1, 2))

    def test_reports(self):
        report_files = [
            os.path.join(dir_name, 'out__00000__update_high.txt'),
            os.path.join(dir_name, 'out__00000__update_low.txt'),
        ]

        combined, obj_val, param_val = pl._combine_files(report_files)
        self.assertIsNotNone(combined)
        self.assertAlmostEqual(obj_val, 12.844)

    def test_axis(self):
        y_min = 60
        y_max = 330
        scale = pl._make_y_axis(y_min, y_max)
        self.assertEqual(scale[0], 60)
        self.assertEqual(scale[-1], 330)

        scale = pl._make_y_axis(y_min, y_max, 5)
        self.assertEqual(scale[0], 0)
        self.assertEqual(scale[-1], 360)

        y_min = 60847326
        y_max = 73425330
        scale = pl._make_y_axis(y_min, y_max)
        self.assertEqual(scale[0], 60000000)
        self.assertEqual(scale[-1], 74000000)

        scale = pl._make_y_axis(0, 12.5015)
        self.assertEqual(scale[0], 0)
        self.assertEqual(scale[-1], 14)

        scale = pl._make_y_axis(12.1, 12.5015, 3)
        self.assertEqual(scale[0], 12)
        self.assertEqual(scale[-1], 12.8)

    def test_creation_and_run(self):
        temp_dir = tempfile.mkdtemp()
        basico.new_model(name='Schaber Example', notes="""
        The example from the supplement, originally described in

        Schaber, J. and Klipp, E. (2011) Model-based inference of biochemical parameters and 
        dynamic properties of microbial signal transduction networks, Curr Opin Biotechnol, 22, 109-
        116 (https://doi.org/10.1016/j.copbio.2010.09.014)

        """)

        basico.add_function('mod. MA',
                            'k * S * T',
                            'irreversible',
                            mapping={'S': 'modifier', 'T': 'substrate'}
                            )

        basico.add_reaction('v1', 'T -> Tp; S', function='mod. MA')
        basico.add_reaction('v2', 'Tp -> T')
        basico.add_reaction('decay', 'S ->', mapping={'k1': 0.5})
        basico.set_species('Tp', initial_concentration=0)
        basico.add_species('TpFit', status='assignment', expression='{[Tp]}/0.5')
        basico.add_experiment('Data set 1', pd.DataFrame(data={
            'Time': [1, 2, 4, 6],
            '[TpFit]': [1, 0.88, 0.39, 0.22],
            # 'sd': [0.09, 0.09, 0.09, 0.09]
        }), file_name=os.path.join(temp_dir, 'data.csv')
        )

        basico.set_fit_parameters([
            {'name': '(v1).k', 'lower': 0, 'upper': 100},
            {'name': '(v2).k1', 'lower': 0, 'upper': 100},
        ])

        temp_file = os.path.join(temp_dir, 'test.cps')
        basico.save_model_and_data(temp_file)

        sol = basico.run_parameter_estimation(method=basico.PE.HOOKE_JEEVES, update_model=True)
        self.assertAlmostEqual(sol['sol'][0], 2.27, places=2)
        self.assertAlmostEqual(sol['sol'][1], 1.43, places=2)

        temp_file = os.path.join(temp_dir, 'test.cps')
        basico.save_model_and_data(temp_file)

        info = pl.prepare_files(temp_file, data_dir=temp_dir, lower_adjustment=0.1, upper_adjustment=10)
        self.assertEqual(info['num_params'], 2)
        self.assertEqual(info['num_data'], 4)
        self.assertAlmostEqual(info['obj'], 0.0059, places=4)
        generated = glob.glob(os.path.join(temp_dir, 'out__*.cps'))
        self.assertIsNotNone(generated)

        info = pl.prepare_files(temp_file, data_dir=temp_dir, lower_adjustment='-2SD', upper_adjustment='+2SD')
        self.assertEqual(info['num_params'], 2)
        self.assertTrue('param_sds' in info)
        self.assertTrue('(v1).k' in info['param_sds'])
        self.assertAlmostEqual(info['param_sds']['(v1).k'], 0.5657, places=2)

        # cleanup
        shutil.rmtree(temp_dir)


if __name__ == '__main__':
    unittest.main()



"""Utility functions for dealing with scan tasks

While usually it is easier to encode scan functionality directly in python code, the utility methods here
make it easy to set up scan tasks as supported by COPASI directly. That way they can be carried out from the
command line version of COPASI, or from the graphical user interface.

"""
import logging

import COPASI
import pandas as pd

import basico
from . import model_io

logger = logging.getLogger(__name__)

def _scan_type_to_name(type):
    names = {
        COPASI.CScanProblem.SCAN_LINEAR: 'scan',
        COPASI.CScanProblem.SCAN_REPEAT: 'repeat',
        COPASI.CScanProblem.SCAN_RANDOM: 'random',
    }
    return names.get(type, 'scan')


def _name_to_scan_type(name):
    types = {
        'scan': COPASI.CScanProblem.SCAN_LINEAR,
        'repeat': COPASI.CScanProblem.SCAN_REPEAT,
        'random': COPASI.CScanProblem.SCAN_RANDOM,
    }
    return types.get(name, COPASI.CScanProblem.SCAN_LINEAR)


def _scan_distribution_type_to_name(type):
    names = {
        0: 'uniform',
        1: 'normal',
        2: 'poisson',
        3: 'gamma',
    }
    return names.get(type, 'uniform')


def _name_to_scan_distribution_type(name):
    values = {
        'uniform': 0,
        'normal': 1,
        'poisson': 2,
        'gamma': 3,
    }
    return values.get(name, 0)


def get_scan_items_frame(**kwargs):
    """Returns all the scan items as pandas DataFrame

    :param kwargs: optional parameters

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return: data frame of scan items
    """
    model = model_io.get_model_from_dict_or_default(kwargs)

    task = model.getTask('Scan')
    assert (isinstance(task, COPASI.CScanTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CScanProblem))

    data = get_scan_items(model=model, problem=problem)

    return pd.DataFrame(data=data)


def get_scan_items(**kwargs):
    """Retrieves the scan items specified on the scan task:

       >>> get_scan_items()
       [
          {
            'type': 'scan',
            'num_steps': 10,
            'log': False,
            'min': 0.5,
            'max': 2.0,
            'item': '(R1).k1',
          }
       ]

    :param kwargs: optional parameters

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return: array of dictionary with the scan items specified
    :rtype: [{}]
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    problem = kwargs.get('problem', model.getTask('Scan').getProblem())

    data = []
    for i in range(problem.getNumberOfScanItems()):
        item = problem.getScanItem(i)
        current = _scan_item_to_dict(item, model)
        data.append(current)
    return data


def _scan_item_to_dict(item, model=None):
    if model is None:
        model = basico.get_current_model()

    if isinstance(item, int):
        problem = model.getTask('Scan').getProblem()
        if item >= problem.getNumberOfScanItems():
            logger.error('No Scan item: {0}'.format(item))
            return None
        item = problem.getScanItem(item)

    assert (isinstance(item, COPASI.CCopasiParameter))
    int_type = item.getParameter('Type').getIntValue()
    type_name = _scan_type_to_name(int_type)
    cn = item.getParameter('Object').getCNValue().getString() if item.getParameter('Object') else None
    num_steps = item.getParameter('Number of steps').getIntValue() if item.getParameter('Number of steps') else None
    min_val = item.getParameter('Minimum').getDblValue() if item.getParameter('Minimum') else None
    max_val = item.getParameter('Maximum').getDblValue() if item.getParameter('Maximum') else None
    log = item.getParameter('log').getBoolValue() if item.getParameter('log') else None
    current = {
        'type': type_name,
        'num_steps': num_steps,
    }

    if log is not None:
        current['log'] = log

    if min_val is not None:
        current['min'] = min_val

    if max_val is not None:
        current['max'] = max_val

    if int_type == COPASI.CScanProblem.SCAN_RANDOM:
        dist = _scan_distribution_type_to_name(item.getParameter('Distribution type').getUIntValue()) \
            if item.getParameter('Distribution type') else None
        current['distribution'] = dist

    elif int_type == COPASI.CScanProblem.SCAN_LINEAR:
        values = item.getParameter('Values').getStringValue() if item.getParameter('Values') else None
        if values:
            values = values.replace(',', ' ')
            values = values.split(' ')
            values = [float(v) for v in values]
        use_values = item.getParameter('Use Values').getBoolValue() if item.getParameter('Use Values') else None
        current['values'] = values
        current['use_values'] = use_values

    if cn:
        obj = model.getObject(COPASI.CCommonName(cn))
        if not obj:
            logger.warning('missing object in scan item: {0}'.format(cn))
        else:
            current['item'] = obj.getObjectDisplayName()
        current['cn'] = cn
    return current


def get_scan_settings(**kwargs):
    """Returns the scan settings as dictionary

        >>> get_scan_settings()
        {
            'update_model': False,
            'scheduled': False,
            'subtask': 'Steady-State',
            'output_during_subtask': False,
            'continue_from_current_state': False,
            'continue_on_error': False,
            'scan_items': [ ... ]
        }

        :param kwargs: optional parameters

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

        :return: dictionary of scan settings
        :rtype: {}
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    task = model.getTask('Scan')
    assert (isinstance(task, COPASI.CScanTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CScanProblem))

    s = {
        'update_model': task.isUpdateModel(),
        'scheduled': task.isScheduled(),
        'subtask': basico.T.from_enum(problem.getSubtask()),
        'output_during_subtask': problem.getOutputInSubtask(),
        'continue_from_current_state': problem.getContinueFromCurrentState(),
        'continue_on_error': problem.getContinueOnError(),
        'scan_items': get_scan_items(model=model, problem=problem)
    }

    return s


def _set_parameter_from_value(parameter, value):
    """Utility function that sets the parameter value

    :param parameter: copasi parameter to change
    :type parameter: COPASI.CCopasiParameter

    :param value: the value to set will be cast to the appropriate type
    :type value: Union[str, float, int, bool]

    """
    if parameter is None:
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_DOUBLE:
        parameter.setDblValue(float(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_INT:
        parameter.setIntValue(int(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_BOOL:
        parameter.setBoolValue(bool(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_STRING \
            or parameter.getType() == COPASI.CCopasiParameter.Type_EXPRESSION:
        parameter.setStringValue(str(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_CN:
        parameter.setCNValue(COPASI.CCommonName(str(value)))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_FILE:
        parameter.setFileValue(str(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_UDOUBLE:
        parameter.setUDblValue(float(value))
        return

    if parameter.getType() == COPASI.CCopasiParameter.Type_UINT:
        parameter.setUIntValue(int(value))


def _set_parameter_from_dict(parameter, values_dict, key):
    """Utility function that sets the given copasi parameter

    :param parameter: copasi parameter to change
    :type parameter: COPASI.CCopasiParameter
    :param values_dict: dictionary of values
    :param key: key that might be in the dictionary
    :return: None
    """
    if not parameter:
        return

    if key not in values_dict:
        return

    _set_parameter_from_value(parameter, values_dict[key])


def add_scan_item(**kwargs):
    """Adds the scan item to the model

      :param kwargs: optional parameters

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

        - | `type (str)`: the type for the scan item can be one of `scan`,
          | `repeat` or `random`. If not specified `scan` will be used.

        - | `cn (str)`: the cn of the element to use in the scan item (use when
          | no suitable display names for the item you are interested in exist)
          | if you specify cn, don't use the `item` optional paramter

        - | `item (str)`: the display name of the item you want to use, for example
          | `[Signal]_0` for the initial concentration of species `Signal`, or
          | `(r1).k` for the local parameter `k` of reaction `r1`.

        - | `values(str or [float])`: if you want to scan over specific values,
          | rather than a range, specify them here e.g.: `[0.1, 0.5, 1, 2]`. Using
          | this parameter also sets `use_values` to `True`

        - | `use_values (bool)`: indicates that the values specified should be used
          | rather than the min / max range

        - | `num_steps (int)`: the number of steps in the range of [min, max] or the
          | number of repeats.

        - | `min (float)`: minimum value for the range or first distribution parameter

        - | `max (float)`: maximum value for the range, or the 2nd distributon parameter

        - | `log (bool)`: boolean indicating that the range should be used logarithmically.

        - | `distribution (str)`: one of `uniform`, `normal`, `poisson` or `gamma`

      :return:
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    problem = kwargs.get('problem', model.getTask('Scan').getProblem())
    assert (isinstance(problem, COPASI.CScanProblem))

    scan_item = kwargs.get('scan_item', kwargs)

    obj = None
    if 'cn' in scan_item:
        obj = model.getObject(COPASI.CCommonName(scan_item['cn']))

    if 'item' in scan_item:
        obj = model.findObjectByDisplayName(scan_item['item'])

    copasi_item = problem.addScanItem(_name_to_scan_type(scan_item['type'] if 'type' in scan_item else 'scan'),
                                      scan_item['num_steps'] if 'num_steps' in scan_item else 0,
                                      obj)
    assert (isinstance(copasi_item, COPASI.CCopasiParameterGroup))

    if 'values' in scan_item:
        values = scan_item['values']
        if isinstance(values, list):
            values = [str(v) for v in values]
            values = ' '.join(values)

        _set_parameter_from_value(copasi_item.getParameter('Values'), values)
        _set_parameter_from_value(copasi_item.getParameter('Use Values'), True)

    _set_parameter_from_dict(copasi_item.getParameter('Use Values'), scan_item, 'use_values')
    _set_parameter_from_dict(copasi_item.getParameter('Minimum'), scan_item, 'min')
    _set_parameter_from_dict(copasi_item.getParameter('Maximum'), scan_item, 'max')
    _set_parameter_from_dict(copasi_item.getParameter('log'), scan_item, 'log')

    if 'distribution' in scan_item:
        _set_parameter_from_value(copasi_item.getParameter('Distribution type'),
                                  _name_to_scan_distribution_type(scan_item['distribution']))


def set_scan_items(scan_items, **kwargs):
    """Replaces the scan items in the task, with the scan items passed in

       If you just wanted to change a specific entry, you could would retrieve
       the current set of scan items, make the change and then set them again:

       >>> scan_items = get_scan_items()

       the list of scan items is returned as array of dictionary items, and
       can be freely modified (say change the number of
       steps of the first item to a new value), or add new entries, remove some

       >>> scan_items[0]['num_steps'] = 20

       Finally a call to scan items, will replace the ones in the model with the
       ones from the array.

       >>> set_scan_items(scan_items)

    :param scan_items: list of dictionaries as returned by :func:`.get_scan_items`.
    :type scan_items: [{}]

    :param kwargs: optional parameters

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    problem = kwargs.get('problem', model.getTask('Scan').getProblem())
    assert (isinstance(problem, COPASI.CScanProblem))

    if isinstance(scan_items, dict):
        scan_items = [scan_items]

    problem.clearScanItems()

    for scan_item in scan_items:
        add_scan_item(scan_item=scan_item, model=model, problem=problem)


def set_scan_settings(**kwargs):
    """Changes the scan settings

    :param kwargs: optional parameters

        -  settings: dictionary with the scan settings
        -  subtask: sub task name
        -  output_during_subtask: boolean, indicating whether ouput should be collected during the subtask execution
        -  continue_from_current_state: boolean indicating, whether the subtask should be reset to initial values (False)
                or not  (True)
        -  continue_on_error: boolean indicating, whether executions should continue, in case one subtask execution failed
           (True) or not.
        -  scan_items: a list of scan items see :func:`.set_scan_items`
        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return:
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    task = model.getTask('Scan')
    assert (isinstance(task, COPASI.CScanTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CScanProblem))

    settings = kwargs.get('settings', kwargs)

    if 'update_model' in settings:
        task.setUpdateModel(settings['update_model'])

    if 'scheduled' in settings:
        task.setScheduled(settings['scheduled'])

    if 'subtask' in settings:
        problem.setSubtask(basico.T.to_enum(settings['subtask']))

    if 'output_during_subtask' in settings:
        problem.setOutputInSubtask(settings['output_during_subtask'])

    if 'continue_from_current_state' in settings:
        problem.setContinueFromCurrentState(settings['continue_from_current_state'])

    if 'continue_on_error' in settings:
        problem.setContinueOnError(settings['continue_on_error'])

    if 'scan_items' in settings:
        set_scan_items(settings['scan_items'], model=model, problem=problem)


def run_scan(**kwargs):
    """Runs the scan task

    :param kwargs: optional parameters

        - | settings: optional dictionary with the scan settings

        - | scan_items: a list of scan items see :func:`.set_scan_items`

        - | output ([(str)]): optional list of cns or display names, of elements
          | to collect in the scan.

        - | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return: None if output is not specified, otherwise the collected output as dataframe.
    :rtype: None or pd.DataFrame
    """
    model = model_io.get_model_from_dict_or_default(kwargs)

    model.getModel().compileIfNecessary()

    if 'settings' in kwargs:
        set_scan_settings(settings=kwargs['settings'], model=model)

    if 'scan_items' in kwargs:
        set_scan_items(kwargs['scan_items'], model=model)

    dh = None
    cols = []
    if 'output' in kwargs:
        dh, cols = basico.task_timecourse.create_data_handler(kwargs['output'], after=kwargs['output'], model=model)
        model.addInterface(dh)

    task = model.getTask('Scan')
    assert (isinstance(task, COPASI.CScanTask))

    if not task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI):
        logger.warning("Couldn't initialize scan task")

    use_initial_values = kwargs.get('use_initial_values', True)
    if not task.processRaw(use_initial_values):
        logger.warning("Couldn't process scan task")

    if dh:
        model.removeInterface(dh)
        data = basico.task_timecourse.get_data_from_data_handler(dh, cols)
        del dh
        dh = None
        return data

    return None



import sys
import unittest
import basico
import basico.compartment_array_tools as cat
import COPASI


class TestArrayTools(unittest.TestCase):
    def test_create_arrays(self):
        dm = basico.load_example('brusselator')
        basico.create_rectangular_array(4, 4, ['X', 'Y'], [0.16, 0.8], delete_template=True)
        compartments = basico.as_dict(basico.get_compartments())
        self.assertEqual(len(compartments), 16)

        names = [c['name'] for c in compartments]
        x_range, y_range, prefixes = cat._split_ranges(names)
        self.assertEqual(x_range, [0, 1, 2, 3])
        self.assertEqual(y_range, [0, 1, 2, 3])
        self.assertEqual(prefixes, ['compartment'])

        # remove some of the generated grid
        cat.delete_compartments(selection=[(0,1), (2,3)])
        compartments = basico.as_dict(basico.get_compartments())
        self.assertEqual(len(compartments), 14)

        names = [c['name'] for c in compartments]
        x_range, y_range, prefixes = cat._split_ranges(names)
        # ensure elements are still sorted
        self.assertEqual(x_range, [0, 1, 2, 3])
        self.assertEqual(y_range, [0, 1, 2, 3])
        self.assertEqual(prefixes, ['compartment'])
        # and that removed elements are not in the list
        self.assertNotIn('compartment[0,1]', names)
        self.assertNotIn('compartment[2,3]', names)
        basico.remove_datamodel(dm)

        dm = basico.load_example('brusselator')
        basico.create_linear_array(4, ['X', 'Y'], [0.16, 0.8], delete_template=True)
        compartments = basico.as_dict(basico.get_compartments())
        self.assertEqual(len(compartments), 4)
        basico.remove_datamodel(dm)

if __name__ == "__main__":
    unittest.main()



import unittest
import basico
import COPASI


class TestOptimization(unittest.TestCase):
    def setUp(self):
        self.dm = basico.new_model(name='Himmelblau')
        self.assertTrue(self.dm is not None)
        self.assertTrue(isinstance(self.dm, COPASI.CDataModel))
        basico.add_parameter('x', initial_value=0)
        basico.add_parameter('y', initial_value=0)
        basico.add_parameter('f', type='assignment',
                             expression='({Values[x].InitialValue}^2+{Values[y].InitialValue}-11)^2+({Values[x].InitialValue}+{Values[y].InitialValue}^2-7)^2')
        
    def test_optitems(self):
        self.assertIsNone(basico.get_opt_parameters())
        template = basico.get_opt_item_template(include_global=True)
        basico.set_opt_parameters(template)
        items = basico.get_opt_parameters()
        self.assertIsNotNone(items)

    def test_settings(self):
        settings = basico.get_opt_settings()
        self.assertIsNotNone(settings)
        self.assertEqual(settings['expression'], '')
        basico.set_objective_function('Values[f].InitialValue', minimize=True)
        settings = basico.get_opt_settings()
        self.assertEqual(settings['expression'], 'Values[f].InitialValue')
        basico.set_opt_settings({'subtask': basico.T.TIME_COURSE})
        self.assertEqual(basico.get_opt_settings()['subtask'], basico.T.TIME_COURSE)

    def test_optimization(self):
        template = basico.get_opt_item_template(include_global=True)
        basico.set_opt_parameters(template)
        basico.set_objective_function('Values[f].InitialValue', minimize=True)
        basico.set_opt_settings(settings={
            'subtask': basico.T.TIME_COURSE,
            'method': {'name': basico.PE.LEVENBERG_MARQUARDT}})
        result = basico.run_optimization()
        self.assertIsNotNone(result)
        stat = basico.get_opt_statistic()
        self.assertIsNotNone(stat)
        self.assertLess(stat['obj'], 1e-6)

    def test_custom_output(self):
        dh, col = basico.create_data_handler([
            # # int values not supported in COPASI 4.35 and below
            # 'CN=Root,Vector=TaskList[Optimization],Problem=Optimization,Reference=Function Evaluations',
            'CN=Root,Vector=TaskList[Optimization],Problem=Optimization,Reference=Best Value'
        ])
        c_list = COPASI.ContainerList()
        c_list.append(self.dm)
        self.assertTrue(dh.compile(c_list))


if __name__ == '__main__':
    unittest.main()



"""This module provides convenience functions for array of compartments.

COPASI can duplicate the current model in either a rectangular or linear manner,
with diffusion reactions added in between the created model. This allows for a
simplified spatial simulation.

This submodule adds functions, to create such an array, to delete the template model.
Additionally some basic plotting functionality is available as well.

Example:

    >>> dm = load_example('brusselator')
    >>> create_rectangular_array(10, 10, ['X', 'Y'], [0.16, 0.8], delete_template=True)
    >>> set_species(['X{compartment[1,1]}',
    ...             'X{compartment[1,2]}',
    ...             'X{compartment[2,1]}',
    ...             'X{compartment[2,2]}'], initial_concentration=10)

    >>> add_event('E0', 'Time > 10', [['X{compartment[1,1]}', '10'],
    ...                              ['X{compartment[1,2]}', '10'],
    ...                              ['X{compartment[2,1]}', '10'],
    ...                              ['X{compartment[2,2]}', '10']])

    >>> data = run_time_course(start_time=0, duration=500)
    >>> animate_rectangular_time_course_as_image(data, metabs=["X", "Y"], min_range=0, max_range=10)


"""
from basico import *

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np
from . import model_io

try:
    from PIL import Image
except ImportError:
    pass

_DEFAULT_Y_LABEL = 'volume y'
_DEFAULT_X_LABEL = 'volume x'
_TITLE_FORMAT_STRING = 'concentrations of "{0}" for each volume at time {1}'


def plot_linear_time_course(data, prefix=None, metab_names=None, shading='gouraud',
                            min_range=np.nan, max_range=np.nan, **kwargs):
    """
    Plots the simulation data on the loaded model assuming that this is in the format
    of an array of compartments as generated by COPASI. This will create a figure for each species

    :param data: timecourse simulation result

    :param prefix: string of compartment prefix to indicate which compartment should be visualized. This is
                   expected to not include the indices, so it would be 'compartment' rather than 'compartment[0]'

    :param metab_names: optional array of metabolite names that should be plotted (defaults to all)

    :param shading: optional shading for the color mesh, defaults to 'gouraud' can also be 'flat'

    :param min_range: optional min range, defaults to NaN, meaning that it is to be the minimum value of the data

    :param max_range: optional max range, defaults to NaN, meaning that it is to be the maximum value of the data

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: array with tuple of figures and their axis
    """

    dm = model_io.get_model_from_dict_or_default(kwargs)
    mod = dm.getModel()
    names = [c.getObjectName() for c in mod.getCompartments()]
    if prefix is not None:
        # prefix was specified so filter down to the compartments needed
        names = [c for c in names if c.startswith(prefix + "[")]
    time = data.index.values

    if metab_names is None:
        metab_names = sorted([col[:col.find('{')] for col in data.columns if '{' + names[0] + '}' in col])

    result = []
    for metab in metab_names:
        arr = [data[metab + '{' + c + '}'] for c in names]
        fig, ax = plt.subplots()
        mesh = ax.pcolormesh(time, np.arange(len(names)), arr, shading=shading)
        vmin, vmax = _get_ranges(arr, min_range, max_range)
        mesh.set_clim(vmin=vmin, vmax=vmax)
        fig.colorbar(mesh, ax=ax)
        ax.set_xlabel('time')
        ax.set_ylabel('volume')
        ax.set_title('concentrations of "%s" for each volume' % metab)
        result.append((fig, ax))

    return result


def _split_ranges(names):
    # type: ([str]) -> ([int], [int], [str])
    x_indices = []
    y_indices = []
    prefixes = []
    for name in names:
        first = name.find('[')
        last = name.rfind(']')
        prefix = name[:first]
        if prefix not in prefixes:
            prefixes.append(prefix)
        coords = name[first + 1: last]
        comma = coords.find(',')
        x = int(coords[:comma])
        y = int(coords[comma + 1:])
        if x not in x_indices:
            x_indices.append(x)
        if y not in y_indices:
            y_indices.append(y)

    x_indices.sort()
    y_indices.sort()
    return x_indices, y_indices, prefixes


def plot_rectangular_time_course(data, times=None, prefix=None, shading='gouraud',
                                 min_range=np.nan, max_range=np.nan, **kwargs):
    """
        Plots the simulation data on the loaded model assuming that this is in the format
        of an array of compartments as generated by COPASI. This will create a figure for each species
        at each of the specified time points.


    :param data: timecourse simulation result

    :param times: optional parameter specifying the times for which to plot the results. If not given, one figure will
                  be created for each output time point in the data

    :param prefix: string of compartment prefix to indicate which compartment should be visualized. This is
                   expected to not include the indices, so it would be 'compartment' rather than 'compartment[0]'

    :param shading: optional shading for the color mesh, defaults to 'gouraud' can also be 'flat'

    :param min_range: optional min range, defaults to NaN, meaning that it is to be the minimum value of the data

    :param max_range: optional max range, defaults to NaN, meaning that it is to be the maximum value of the data

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: array with tuple of figures and their axis
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    mod = dm.getModel()
    names = [c.getObjectName() for c in mod.getCompartments()]
    x_range, y_range, prefixes = _split_ranges(names)
    time = data.index.values
    if times is None:
        times = time

    if prefix is None:
        prefix = prefixes[0]

    metab_names = sorted([col[:col.find('{')] for col in data.columns if '{' + names[0] + '}' in col])

    result = []
    for i in range(len(data)):
        cur = data.iloc[i]
        t = time[i]
        if t not in times:
            # skip all times we don't need
            continue

        for metab in metab_names:
            arr = _extract_metabolite_data(cur, metab, prefix, x_range, y_range)
            fig, ax = plt.subplots()
            mesh = ax.pcolormesh(x_range, y_range, arr, shading=shading)
            vmin, vmax = _get_ranges(arr, min_range, max_range)
            mesh.set_clim(vmin=vmin, vmax=vmax)
            fig.colorbar(mesh, ax=ax)
            ax.set_xlabel(_DEFAULT_X_LABEL)
            ax.set_ylabel(_DEFAULT_Y_LABEL)
            ax.set_title(_TITLE_FORMAT_STRING.format(metab, t))
            result.append((fig, ax))

    return result


def _extract_metabolite_data(cur, metab, prefix, x_range, y_range):
    num_x = len(x_range)
    num_y = len(y_range)

    arr = np.zeros((num_x, num_y)) * np.nan
    for x in range(num_x):
        for y in range(num_y):
            metab_name = metab + '{' + prefix + '[{0},{1}]'.format(x_range[x], y_range[y]) + '}'
            if metab_name in cur:
                arr[x, y] = cur[metab_name]
    return arr


def animate_rectangular_time_course_as_image(data, metabs=None, prefix=None,
                                             min_range=np.nan, max_range=np.nan, filename=None, **kwargs):
    """Plots the data as image.

    :param data: data frame with results to plot

    :param metabs:

    :param prefix: string of compartment prefix to indicate which compartment should be visualized. This is
                   expected to not include the indices, so it would be 'compartment' rather than 'compartment[0]'


    :param min_range: optional min range, defaults to NaN, meaning that it is to be the minimum value of the data

    :param max_range: optional max range, defaults to NaN, meaning that it is to be the maximum value of the data

    :param filename: optional filename to safe the image as

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: FuncAnimation result
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    mod = dm.getModel()
    names = [c.getObjectName() for c in mod.getCompartments()]
    x_range, y_range, prefixes = _split_ranges(names)
    time = data.index.values

    metab_names = sorted([col[:col.find('{')] for col in data.columns if '{' + names[0] + '}' in col])
    if metabs is None:
        metabs = metab_names

    if prefix is None:
        prefix = prefixes[0]

    metab_data = []

    vmax = -np.inf if np.isnan(max_range) else max_range
    vmin = np.inf if np.isnan(min_range) else min_range

    cur = data.iloc[0]
    for i in range(len(metabs)):
        metab = metabs[i]
        arr = _extract_metabolite_data(cur, metab, prefix, x_range, y_range)
        vmin, vmax = _get_ranges(arr, min_range, max_range, vmin, vmax)
        metab_data.append(arr)

    img = _create_image(metab_data, vmax)

    fig, ax = plt.subplots()
    imgplot = ax.imshow(img)
    ax.set_xlabel(_DEFAULT_X_LABEL)
    ax.set_ylabel(_DEFAULT_Y_LABEL)
    ax.set_title(_TITLE_FORMAT_STRING.format(metabs, time[0]))

    def _plot_ith_set(time_index):
        cur = data.iloc[time_index]
        vmax = -np.inf if np.isnan(max_range) else max_range
        vmin = np.inf if np.isnan(min_range) else min_range
        metab_data = []
        for i in range(len(metabs)):
            metab = metabs[i]
            arr = _extract_metabolite_data(cur, metab, prefix, x_range, y_range)
            vmin, vmax = _get_ranges(arr, min_range, max_range, vmin, vmax)
            metab_data.append(arr)

        imgplot.set_data(_create_image(metab_data, vmax))
        ax.set_title(_TITLE_FORMAT_STRING.format(metabs, time[time_index]))
        return [imgplot]

    anim = FuncAnimation(
        fig, _plot_ith_set, interval=100, frames=len(time) - 1)

    if filename:
        anim.save(filename)

    return anim


def _create_image(metab_data, vmax):
    img = Image.new('RGB', metab_data[0].shape, color=(73, 109, 137))
    for x in range(img.size[0]):
        for y in range(img.size[1]):
            color = [0, 0, 0]
            for i in range(len(metab_data)):
                color[i] = int(((metab_data[i][x, y] / vmax) * 255))
            img.putpixel((x, y), tuple(color))
    return img


def _get_ranges(arr, min_range=np.nan, max_range=np.nan, vmin=np.inf, vmax=-np.inf):
    vmin = min(vmin, np.min(arr)) if np.isnan(min_range) else min_range
    vmax = max(vmax, np.max(arr)) if np.isnan(max_range) else max_range
    return vmin, vmax


def animate_rectangular_time_course(data, metab=None, prefix=None, shading='gouraud',
                                    min_range=np.nan, max_range=np.nan, filename=None, **kwargs):
    """
        Plots the simulation data on the loaded model assuming that this is in the format
        of an array of compartments as generated by COPASI. This will create a figure for each species
        at each of the specified time points.


    :param data: timecourse simulation result
    :param metab: optional parameter specifying the species to animate. If not given, one the first metab will be chosen
    :param prefix: string of compartment prefix to indicate which compartment should be visualized. This is
                   expected to not include the indices, so it would be 'compartment' rather than 'compartment[0]'
    :param shading: optional shading for the color mesh, defaults to 'gouraud' can also be 'flat'
    :param min_range: optional min range, defaults to NaN, meaning that it is to be the minimum value of the data
    :param max_range: optional max range, defaults to NaN, meaning that it is to be the maximum value of the data
    :param filename: optional filename to a file to which to save the animation to

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: the FuncAnimation constructed
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    mod = dm.getModel()
    names = [c.getObjectName() for c in mod.getCompartments()]
    x_range, y_range, prefixes = _split_ranges(names)
    time = data.index.values

    metab_names = sorted([col[:col.find('{')] for col in data.columns if '{' + names[0] + '}' in col])
    if metab is None:
        metab = metab_names[0]

    if prefix is None:
        prefix = prefixes[0]

    fig, ax = plt.subplots()
    arr = _extract_metabolite_data(data.iloc[0], metab, prefix, x_range, y_range)
    mesh = ax.pcolormesh(x_range, y_range, arr, shading=shading)
    vmin, vmax = _get_ranges(arr, min_range, max_range)
    mesh.set_clim(vmin=vmin, vmax=vmax)
    fig.colorbar(mesh, ax=ax)
    ax.set_xlabel(_DEFAULT_X_LABEL)
    ax.set_ylabel(_DEFAULT_Y_LABEL)
    ax.set_title(_TITLE_FORMAT_STRING.format(metab, 0))

    def _plot_ith_set(i):
        arr = _extract_metabolite_data(data.iloc[i], metab, prefix, x_range, y_range)
        vmin, vmax = _get_ranges(arr, min_range, max_range)
        mesh.set_array(arr.ravel())
        mesh.set_clim(vmin=vmin, vmax=vmax)
        ax.set_title(_TITLE_FORMAT_STRING.format(metab, time[i]))

    anim = FuncAnimation(
        fig, _plot_ith_set, interval=100, frames=len(time) - 1)

    if filename:
        anim.save(filename)

    return anim


# def _get_selection(x_range, y_range):
#     import sys
#     from PyQt5.QtWidgets import QApplication, QGridLayout, QDialog, QPushButton

#     app = QApplication(sys.argv)

#     layout = QGridLayout()

#     w = QDialog()
#     w.resize(250, 150)
#     w.move(300, 300)
#     w.setWindowTitle("Toggle off elements you don't want")

#     buttons = {}

#     for i in x_range:
#         for j in y_range:
#             # keep a reference to the buttons
#             buttons[(i, j)] = QPushButton('.', w)
#             buttons[(i, j)].setCheckable(True)
#             buttons[(i, j)].setMaximumSize(10, 10)
#             # add to the layout
#             layout.addWidget(buttons[(i, j)], i, j)

#     w.setLayout(layout)
#     w.exec()

#     result = []
#     # get selection
#     for coord in buttons:
#         if buttons[coord].isChecked():
#             result.append(coord)
#     return result


# def _delete_compartments(dm):
#     mod = dm.getModel()
#     assert (isinstance(mod, COPASI.CModel))
#     names = [c.getObjectName() for c in mod.getCompartments()]
#     x_range, y_range, prefixes = _split_ranges(names)
#     to_delete = _get_selection(x_range, y_range)
#     delete_compartments(dm, to_delete)


def delete_compartments(selection, **kwargs):
    """
    utility function for deleting a selection of compartments from the datamodel. This will also delete
    the species and reactions included.

    :param selection: an array of tuples of indices at which the compartments should be deleted

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    mod = dm.getModel()
    assert (isinstance(mod, COPASI.CModel))
    names = [c.getObjectName() for c in mod.getCompartments()]
    x_range, y_range, prefixes = _split_ranges(names)
    for prefix in prefixes:
        for coord in selection:
            c = mod.getCompartment(prefix + "[" + str(coord[0]) + "," + str(coord[1]) + "]")
            if c is None:
                continue
            key = c.getKey()
            del c
            mod.removeCompartment(key, True)
            mod.forceCompile()


def _create_array(dm, num_steps_x, num_steps_y, linear=True, species=None, diffusion_coefficients=None,
                  compartment_names=None, delete_template=False):
    if not species:
        return
    model = dm.getModel()
    metabs = []
    model_elements = COPASI.CModelExpansion_SetOfModelElements()
    metab_set = COPASI.DataObjectSet()
    keys = []
    names = []

    for name in species:
        metab = dm.findObjectByDisplayName(name)
        if metab is None:
            continue
        metabs.append(metab)
        metab_set.append(metab)
        compartment_name = metab.getCompartment().getObjectName()
        if compartment_names is None:
            compartment_names = [compartment_name]
        names.append('diff_{0}_{1}'.format(compartment_name, metab.getObjectName()))

    for name in compartment_names:
        comp = dm.findObjectByDisplayName(name)
        if comp is None:
            comp = model.getCompartment(name)
            if comp is None:
                continue
        model_elements.addCompartment(comp)
        keys.append(comp.getKey())

    me = COPASI.CModelExpansion(model)
    model_elements.fillDependencies(model)
    if linear:
        me.createLinearArray(model_elements, num_steps_x, metab_set)
    else:
        me.createRectangularArray(model_elements, num_steps_x, num_steps_y, metab_set)

    if delete_template:
        for key in keys:
            model.removeCompartment(key, True)
            model.forceCompile()

    if diffusion_coefficients is not None:
        for i in range(min(len(names), len(diffusion_coefficients))):
            p = model.getModelValue(names[i])
            p.setInitialValue(diffusion_coefficients[i])


def create_linear_array(num_steps, species=None, diffusion_coefficients=None, compartment_names=None,
                        delete_template=False, **kwargs):
    """
    Utility function to create a linear duplicating the specified species, their reactions
    in the given compartments and created diffusion reactions between the newly created array

    :param num_steps: the number of steps to create

    :param species: array of species names, that should be diffusing between compartments

    :param compartment_names: optional compartment names (will default to the compartment the species is in)

    :param diffusion_coefficients: optional array of diffusion coefficients in the same order as the species
           (otherwise they will be set to 1)

    :param delete_template: if True, the original template model in the specified compartment will be deleted.

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    _create_array(dm, num_steps, 0, linear=True, species=species,
                  compartment_names=compartment_names,
                  diffusion_coefficients=diffusion_coefficients,
                  delete_template=delete_template)


def create_rectangular_array(num_steps_x, num_steps_y, species=None, diffusion_coefficients=None,
                             compartment_names=None, delete_template=False, **kwargs):
    """
    Utility function to create a rectangular array duplicating the specified species, their reactions
    in the given compartments and created diffusion reactions between the newly created array

    :param num_steps_x: the number of compartments to create along the x direction

    :param num_steps_y: the number of compartments to create along the y direction

    :param species: array of species names, that should be diffusing between compartments

    :param compartment_names: optional compartment names (will default to the compartment the species is in)

    :param diffusion_coefficients: optional array of diffusion coefficients in the same order as the species
           (otherwise they will be set to 1)

    :param delete_template: if True, the original template model in the specified compartment will be deleted.

    :param kwargs:

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)
    _create_array(dm, num_steps_x, num_steps_y, linear=False, species=species,
                  diffusion_coefficients=diffusion_coefficients,
                  compartment_names=compartment_names,
                  delete_template=delete_template)



"""This module encapsulates methods for running time course simulations.

main method provided is the :func:`run_time_course` method, that will
simulate the given model (or the current :func:`.get_current_model`).

Examples:

    To run a time course for the duration of 10 time units use

    >>> run_time_course(10)

    To run a time course for the duration of 10 time units, in 50 simulation steps use

    >>> run_time_course(10, 50)

    To run a time course from 0, for the duration of 10 time units, in 50 simulation steps use:

    >>> run_time_course(0, 10, 50)

    all parameters can also be given as key value pairs.

"""


import COPASI
import pandas as pd

import basico
from . import model_io
from . import model_info
from .callbacks import get_default_handler

import pandas
import numpy
import logging

logger = logging.getLogger(__name__)

def __build_result_from_ts(time_series, use_concentrations=True, use_sbml_id=False, model=None):
    # type: (COPASI.CTimeSeries) -> pandas.DataFrame
    col_count = time_series.getNumVariables()
    row_count = time_series.getRecordedSteps()

    if use_sbml_id and model is None:
        model = model_io.get_current_model()

    column_names = []
    column_keys = []
    for i in range(col_count):
        column_keys.append(time_series.getKey(i))
        name = time_series.getTitle(i)

        if use_sbml_id and name != 'Time':
            sbml_id = time_series.getSBMLId(i, model)
            if sbml_id:
                name = sbml_id

        column_names.append(name)

    concentrations = numpy.empty([row_count, col_count])
    for i in range(row_count):
        for j in range(col_count):
            if use_concentrations: 
                concentrations[i, j] = time_series.getConcentrationData(i, j)
            else:
                concentrations[i, j] = time_series.getData(i, j)

    df = pandas.DataFrame(data=concentrations, columns=column_names)
    df = df.set_index('Time')

    return df


def __method_name_to_type(method_name):
    methods = {
        'deterministic': COPASI.CTaskEnum.Method_deterministic,
        'lsoda': COPASI.CTaskEnum.Method_deterministic,
        'hybrid': COPASI.CTaskEnum.Method_hybrid,
        'hybridode45': COPASI.CTaskEnum.Method_hybridODE45,
        'hybridlsoda': COPASI.CTaskEnum.Method_hybridLSODA,
        'adaptivesa': COPASI.CTaskEnum.Method_adaptiveSA,
        'tauleap': COPASI.CTaskEnum.Method_tauLeap,
        'stochastic': COPASI.CTaskEnum.Method_stochastic,
        'directmethod': COPASI.CTaskEnum.Method_directMethod,
        'radau5': COPASI.CTaskEnum.Method_RADAU5,
        'sde': COPASI.CTaskEnum.Method_stochasticRunkeKuttaRI5,
    }
    return methods.get(method_name.lower(), COPASI.CTaskEnum.Method_deterministic)


def run_time_course_with_output(output_selection, *args, **kwargs):
    """Simulates the current model, returning only the data specified in the output_selection array

    :param output_selection: selection of elements to return, for example ['Time', '[ATP]', 'ATP.Rate'] to
          return the time column, ATP concentration and the rate of change of ATP. The strings can be either
          the Display names as can be found in COPASI, or the CN's of the elements.

    :param args: positional arguments

     * 1 argument: the duration to simulate the model
     * 2 arguments: the duration and number of steps to take
     * 3 arguments: start time, duration, number of steps

    :param kwargs: additional arguments

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

     - `use_initial_values` (bool): whether to use initial values

     - `scheduled` (bool): sets whether the task is scheduled or not

     - `update_model` (bool): sets whether the model should be updated, or reset to initial conditions.

     - | `method` (str): sets the simulation method to use (otherwise the previously set method will be used)
       | support methods:
       |
       |   * `deterministic` / `lsoda`: the LSODA implementation
       |   * `stochastic`: the Gibson & Bruck Gillespie implementation
       |   * `directMethod`: Gillespie Direct Method
       |   * others: `hybrid`, `hybridode45`, `hybridlsoda`, `adaptivesa`, `tauleap`, `radau5`, `sde`

     - `duration` (float): the duration in time units for how long to simulate

     - `automatic` (bool): whether to use automatic determined steps (True), or the specified interval / number of steps

     - `output_event` (bool): if true, output will be collected at the time a discrete event occurs.

     - | `values` ([float]): if given, output will only returned at the output points specified
       |                     for example use `values=[0, 1, 4]` to return output only for those three times

     - | `start_time` (float): the output start time. If the model is not at that start time, a simulation
       |  will be performed in one step, to reach it before starting to collect output.

     - | `step_number` or `intervals` (int): the number of output steps. (will only be used if `automatic`
       | or `stepsize` is not used.

     - | `stepsize` (float): the output step size (will only be used if `automatic` is False).

     - | `seed` (int): set the seed that will be used if `use_seed` is true, using this stochastic trajectories can
       | be repeated

     - | 'use_seed' (bool): if true, the specified seed will be used.

     - | `a_tol` (float): the absolute tolerance to be used

     - | `r_tol` (float): the relative tolerance to be used

     - | `max_steps` (int): the maximum number of internal steps the integrator is allowed to use.

    """
    model = model_io.get_model_from_dict_or_default(kwargs)

    model.getModel().compileIfNecessary()

    dh, columns = create_data_handler(output_selection, model=model)

    task, use_initial_values = _setup_timecourse(args, kwargs)

    model.addInterface(dh)
    result = task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI)
    task.setCallBack(get_default_handler())
    if not result:
        logger.error("Error while initializing the simulation: " +
        COPASI.CCopasiMessage.getLastMessage().getText())
    else:
        result = task.processRaw(use_initial_values)
        if not result:
            logger.error("Error while running the simulation: " +
            COPASI.CCopasiMessage.getLastMessage().getText())
    task.setCallBack(COPASI.CProcessReport())
    df = get_data_from_data_handler(dh, columns)
    model.removeInterface(dh)

    return df


def get_data_from_data_handler(dh, columns):
    data = []
    for i in range(dh.getNumRowsDuring()):
        row = dh.getNthRow(i)
        current_row = []
        for element in row:
            current_row.append(element)
        data.append(current_row)
    df = pd.DataFrame(data=data, columns=columns)
    return df


def create_data_handler(output_selection, during=None, after=None, before=None, model=None):
    """Creates an output handler for the given selection

    :param output_selection: list of display names or cns, of elements to capture
    :type output_selection: [str]

    :param during: optional list of elements from the output selection, that should be collected
                   during the run of the task
    :type during: [str]

    :param after: optional list of elements from the output selection, that should be collected
                   after the run of the task
    :type after: [str]

    :param before: optional list of elements from the output selection, that should be collected
                  before the run of the task
    :type before: [str]

    :param model: the model in which to resolve the display names

    :return: tuple of the data handler from which to retrieve output later, and their columns
    :rtype: (COPASI.CDataHandler, [])

    """
    if model is None:
        model = basico.get_current_model()

    dh = COPASI.CDataHandler()
    columns = []
    for name in output_selection:
        if name.startswith('CN='):
            obj = model.getObject(COPASI.CCommonName(name))
            if not obj:
                logger.warning('no object for cn {0}'.format(name))
                continue
            cn = name
            columns.append(obj.getObjectDisplayName())
        else:
            obj = model.findObjectByDisplayName(name)

            if not obj:
                logger.warning('no object for name {0}'.format(name))
                continue

            if obj.getObjectType() != 'Reference':
                obj = obj.getValueReference()

            cn = obj.getCN().getString()
            columns.append(name)

        if during is None or (during is not None and name in during):
            dh.addDuringName(COPASI.CRegisteredCommonName(cn))

        if after and name in after:
            dh.addAfterName(COPASI.CRegisteredCommonName(cn))

        if before and name in before:
            dh.addAfterName(COPASI.CRegisteredCommonName(cn))
    return dh, columns


def run_time_course(*args, **kwargs):
    """Simulates the current or given model, returning a data frame with the results

    :param args: positional arguments

     * 1 argument: the duration to simulate the model
     * 2 arguments: the duration and number of steps to take
     * 3 arguments: start time, duration, number of steps

    :param kwargs: additional arguments

     - | `model`: to specify the data model to be used (if not specified
       | the one from :func:`.get_current_model` will be taken)

     - `use_initial_values` (bool): whether to use initial values

     - `scheduled` (bool): sets whether the task is scheduled or not

     - `update_model` (bool): sets whether the model should be updated, or reset to initial conditions.

     - | `method` (str): sets the simulation method to use (otherwise the previously set method will be used)
       | support methods:
       |
       |   * `deterministic` / `lsoda`: the LSODA implementation
       |   * `stochastic`: the Gibson & Bruck Gillespie implementation
       |   * `directMethod`: Gillespie Direct Method
       |   * others: `hybrid`, `hybridode45`, `hybridlsoda`, `adaptivesa`, `tauleap`, `radau5`, `sde`

     - `duration` (float): the duration in time units for how long to simulate

     - `automatic` (bool): whether to use automatic determined steps (True), or the specified interval / number of steps

     - `output_event` (bool): if true, output will be collected at the time a discrete event occurs.

     - | `values` ([float]): if given, output will only returned at the output points specified
       |                     for example use `values=[0, 1, 4]` to return output only for those three times

     - | `start_time` (float): the output start time. If the model is not at that start time, a simulation
       |  will be performed in one step, to reach it before starting to collect output.

     - | `step_number` or `intervals` (int): the number of output steps. (will only be used if `automatic`
       | or `stepsize` is not used.

     - | `stepsize` (float): the output step size (will only be used if `automatic` is False).

     - | `seed` (int): set the seed that will be used if `use_seed` is true, using this stochastic trajectories can
       | be repeated

     - | 'use_seed' (bool): if true, the specified seed will be used.

     - | `a_tol` (float): the absolute tolerance to be used

     - | `r_tol` (float): the relative tolerance to be used

     - | `max_steps` (int): the maximum number of internal steps the integrator is allowed to use.

     - | `use_concentrations` (bool): whether to return just the concentrations (default)

     - | `use_numbers` (bool): return all elements collected

    :return: data frame with simulation results
    :rtype: pandas.DataFrame
    """
    model = model_io.get_model_from_dict_or_default(kwargs)

    model.getModel().compileIfNecessary()

    task, use_initial_values = _setup_timecourse(args, kwargs)

    result = task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI)
    if not result:
        logger.error("Error while initializing the simulation: " +
        COPASI.CCopasiMessage.getLastMessage().getText())
    else:
        task.setCallBack(get_default_handler())
        result = task.processRaw(use_initial_values)
        if not result:
            logger.error("Error while running the simulation: " +
            COPASI.CCopasiMessage.getLastMessage().getText())

    use_concentrations = kwargs.get('use_concentrations', True)
    if 'use_numbers' in kwargs and kwargs['use_numbers']:
        use_concentrations = False

    use_sbml_id = kwargs.get('use_sbml_id', False)
    
    return __build_result_from_ts(task.getTimeSeries(), use_concentrations, use_sbml_id, model)


def _setup_timecourse(args, kwargs):
    model = model_io.get_model_from_dict_or_default(kwargs)
    num_args = len(args)
    use_initial_values = kwargs.get('use_initial_values', True)
    task = model.getTask('Time-Course')
    assert (isinstance(task, COPASI.CTrajectoryTask))
    if 'scheduled' in kwargs:
        task.setScheduled(kwargs['scheduled'])
    if 'update_model' in kwargs:
        task.setUpdateModel(kwargs['update_model'])
    if 'method' in kwargs:
        task.setMethodType(__method_name_to_type(kwargs['method']))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CTrajectoryProblem))
    if 'duration' in kwargs:
        problem.setDuration(float(kwargs['duration']))
        problem.setUseValues(False)
    if 'automatic' in kwargs:
        problem.setAutomaticStepSize(bool(kwargs['automatic']))
    if 'output_event' in kwargs:
        problem.setOutputEvent(bool(kwargs['output_event']))
    if 'start_time' in kwargs:
        problem.setOutputStartTime(float(kwargs['start_time']))
    if 'step_number' in kwargs:
        problem.setStepNumber(int(kwargs['step_number']))
    if 'intervals' in kwargs:
        problem.setStepNumber(int(kwargs['intervals']))
    if 'stepsize' in kwargs:
        problem.setStepSize(float(kwargs['stepsize']))
    if 'values' in kwargs:
        vals = kwargs['values']
        if type(vals) != str:
            new_vals = ''
            for val in vals:
                new_vals += ' ' + str(val)
            vals = new_vals.strip()
        problem.setValues(vals)
        problem.setUseValues(True)
    if 'use_values' in kwargs:
        problem.setUseValues(bool(kwargs['use_values']))
    if num_args == 3:
        problem.setOutputStartTime(float(args[0]))
        problem.setDuration(float(args[1]))
        problem.setStepNumber(int(args[2]))
        problem.setUseValues(False)
    elif num_args == 2:
        problem.setDuration(float(args[0]))
        problem.setStepNumber(int(args[1]))
        problem.setUseValues(False)
    elif num_args > 0:
        problem.setDuration(float(args[0]))
    problem.setTimeSeriesRequested(True)
    method = task.getMethod()
    if 'seed' in kwargs and method.getParameter('Random Seed'):
        method.getParameter('Random Seed').setIntValue(int(kwargs['seed']))
    if 'use_seed' in kwargs and method.getParameter('Random Seed'):
        method.getParameter('Use Random Seed').setBoolValue(bool(kwargs['use_seed']))
    if 'a_tol' in kwargs and method.getParameter('Absolute Tolerance'):
        method.getParameter('Absolute Tolerance').setDblValue(float(kwargs['a_tol']))
    if 'r_tol' in kwargs and method.getParameter('Relative Tolerance'):
        method.getParameter('Relative Tolerance').setDblValue(float(kwargs['r_tol']))
    if 'max_steps' in kwargs and method.getParameter('Max Internal Steps'):
        method.getParameter('Max Internal Steps').setIntValue(int(kwargs['max_steps']))
    if 'settings' in kwargs:
        model_info.set_task_settings(task, kwargs['settings'])
    return task, use_initial_values



""" Utility functions for generating profile likelihood plots

   This module contains utility functions for generating scans
   around the current solution of a parameter estimation. The method
   is described in detail in:

   Schaber J (2012) Easy parameter identifiability analysis with COPASI.
   Biosystems 110:183–185 http://dx.doi.org/10.1016/j.biosystems.2012.09.003

"""
import math

import basico
import numpy as np
import pandas as pd
import scipy.stats as stats
import sys
import os

from basico.processing import process_files
import tempfile
import logging
import shutil
import COPASI
import yaml

try:
    from . import model_io
except ImportError:
    from basico import model_io

COPASI_SE = 'CopasiSE'
logger = logging.getLogger(__name__)


def _get_scan_files(data_dir):
    """Collects any of the scans generated from the given data directory

    :param data_dir: the data directory
    :return: list of absolute filenames from the directory
    """
    files = []
    for filename in os.listdir(data_dir):
        if not (filename.endswith('_update_high.cps') or filename.endswith('_update_low.cps')):
            continue
        files.append(os.path.abspath(os.path.join(data_dir, filename)))
    return files


def process_dir(data_dir, pool_size=4, copasi_se=COPASI_SE, max_time=None):
    """Executes all generated copasi files from the given directory in a multiprocessing pool

    :param data_dir: directory containing high / low files generated
    :param pool_size: pool size for the multiprocessing pool that will be created (defaults to 4)
    :param copasi_se: path to the copasi SE executable to execute
    :param max_time: maximum time to allow for the execution of each file in seconds (defaults to None for no maximum
    """
    files = _get_scan_files(data_dir)
    process_files(files, pool_size, copasi_se, max_time)


def plot_data(data_dir, problem_size=None, **kwargs):
    """Plots all files from the given data directory

    :param data_dir: data directory containing the report files

    :param problem_size: tuple containing the number of parameters (m) and number of data points (n)

    :param kwargs: additional keyword arguments to pass to the plot function

        - scale_mode: scale_mode to use for the y-axis (defaults to None for automatic scaling). If scale_mode
          is a float, we compute a scale between objective value and threshold with the given steps.
          If scale_mode is a dict, we use the given dict as the scale for `set_ylim`.

        - obj_val: objective value to use for the plot (defaults to None for using the value from the info file)

        - alpha: alpha value to use for the calculation of thresholds (defaults to 0.05)

        - thresholds: list of thresholds to calculate and display. if not specified, we plot the thresholds
                      for the 95% interval, and 68% interval according to obj*(1+chi2/(n-m)). Valid values are:

            - 'default_68: threshold for 68%
            - 'default_95: threshold for 95%
            - 'schaber_chi2_1p': threshold for one parameter according to Schaber using chi2 with 1 parameter
            - 'schaber_chi2_p': threshold for one parameter according to Schaber using chi2 with p parameters
            - 'schaber_fratio_p': threshold for one parameter according to Schaber using fratio with p parameters
            - `donaldson_fratio_1p`: threshold for one parameter according to Donaldson using fratio with 1 parameter

    :return: list of Axes created
    :rtype: List[matplotlib.axes.Axes]
    """

    if os.path.exists(os.path.join(data_dir, '.info.yaml')):
        with open(os.path.join(data_dir, '.info.yaml'), 'r', encoding='utf8') as f:
            info = yaml.safe_load(f)
        param_sds = info.get('param_sds', {})
    else:
        info = None
        param_sds = {}

    file_map = _get_report_filemap(data_dir)
    plots = []
    first = True
    for entry in file_map:
        df, obj_val, param_val = _combine_files(file_map[entry])
        if 'obj_val' in kwargs:
            obj_val = kwargs['obj_val']
        elif info is not None:
            obj_val = info['obj']

        if not problem_size and info is not None:
            problem_size = (info['num_params'], info['num_data'])
        param_name = df.index.name
        ax = df.plot(legend=False)
        ax.set_ylabel('obj')
        ax.axvline(param_val, color='silver', ls='dotted')
        if param_name in param_sds:
            ax.axvline(param_val + param_sds[param_name], color='lightblue', ls='dotted')
            ax.axvline(param_val + 2 * param_sds[param_name], color='lavender', ls='dotted')
            ax.axvline(param_val - param_sds[param_name], color='lightblue', ls='dotted')
            ax.axvline(param_val - 2 * param_sds[param_name], color='lavender', ls='dotted')
        ax.axhline(obj_val, color='silver', ls='dotted')
        thresholds_to_plot = kwargs.get('thresholds', ['default_95', 'default_68'])
        if problem_size:
            m, n = problem_size
            scale_mode = kwargs.get('scale_mode', 3)
            log_level = kwargs.get('log_level', logger.INFO)
            if type(log_level) is str:
                log_level = logger.getLevelName(log_level.upper())

            if scale_mode is not None:
                if type(scale_mode) is float:
                    scale = _make_y_axis(obj_val, threshold, scale_mode)
                    ax.set_ylim(top=scale[-1], bottom=scale[0])

                if type(scale_mode) is dict:
                    ax.set_ylim(**scale_mode)

            # threshold for COPASI chi2 alpha is 68% (sahle)
            if 'default_68' in thresholds_to_plot:
                c0 = stats.chi2.isf(0.32, 1, loc=0, scale=1)
                threshold = obj_val * (1 + c0/(n-m))
                if first:
                    logger.log(log_level, f'blue, dashed : COPASI threshold {threshold} for alpha=0.32')
                ax.axhline(threshold, color='blue', ls='dashed')

            # for remaining thresholds we use alpha = 0.05 unless otherwise specified
            alpha = kwargs.get('alpha', 0.05)

            # threshold for COPASI chi2 alpha is 95% (sahle)
            if 'default_95' in thresholds_to_plot:
                c0 = stats.chi2.isf(alpha, 1, loc=0, scale=1)
                threshold = obj_val * (1 + c0/(n-m))
                if first:
                    logger.log(log_level, f'blue, dotted : COPASI threshold {threshold} for alpha={alpha}')
                ax.axhline(threshold, color='blue', ls='dotted')

            # estimating chi-square value fitting one parameter (schaber)
            if 'schaber_chi2_1p' in thresholds_to_plot:
                c1 = stats.chi2.isf(alpha, 1, loc=0, scale=1)
                t1 = obj_val * math.exp(c1 / n)
                if first:
                    logger.log(log_level, f'green, dashed: Schaber threshold {t1} for alpha={alpha}, 1 parameter, chi2')
                ax.axhline(y=t1, color='green', linestyle='dashed')

            # estimating chi-square value fitting m parameters (schaber)
            if 'schaber_chi2_p' in thresholds_to_plot:
                c2 = stats.chi2.isf(alpha, m, loc=0, scale=1)
                t2 = obj_val * math.exp(c2 / n)
                if first:
                    logger.log(log_level, f'green, dotted: Schaber threshold {t2} for alpha={alpha}, m parameter, chi2')
                ax.axhline(y=t2, color='green', linestyle='dotted')

            # estimating fratio value fitting m parameters (schaber)
            if 'schaber_fratio_p' in thresholds_to_plot:
                c3 = stats.f.isf(alpha, m, n-m, loc=0, scale=1)
                t3 = obj_val * (1 + (m / (n-m)) * c3)
                if first:
                    logger.log(log_level, f'red: Schaber threshold {t3} for alpha={alpha}, m parameter, Fratio')
                ax.axhline(y=t3, color='red', linestyle='dotted')

            # estimating chi-square value fitting 1 parameter (donaldson)
            if 'donaldson_fratio_1p' in thresholds_to_plot:
                c4 = stats.f.isf(alpha, 1, n-m, loc=0, scale=1)
                t4 = obj_val * (1 + c4/(n-m))
                if first:
                    logger.log(log_level, f'orange: Donaldson threshold {t4} for alpha={alpha}, 1 parameter, Fratio')
                ax.axhline(y=t4, color='orange', linestyle='dashed')

            first = False

        ax.set_title(f'{df.index.name} obj={obj_val}, value={param_val}')
        plots.append(ax)

    return plots


def _make_y_axis(y_min, y_max, ticks=10):
    """

    This routine creates the Y axis values for a graph.

    Calculate Min amd Max graphical labels and graph
    increments.  The number of ticks defaults to
    10 which is the SUGGESTED value.  Any tick value
    entered is used as a suggested value which is
    adjusted to be a 'pretty' value.

    Output will be an array of the Y axis values that
    encompass the Y values.

    translated from this SO post:
    https://stackoverflow.com/questions/326679/choosing-an-attractive-linear-scale-for-a-graphs-y-axis/9007526#9007526

    :param y_min: min value of y-axis
    :param y_max: max value of y-axis
    :param ticks: number of ticks to return (defaults to 10)
    :return: array with the suggested scale
    """

    result = []

    # If y_min and y_max are identical, then
    # adjust the y_min and y_max values to actually
    # make a graph. Also avoids division by zero errors.

    if y_min == y_max:
        y_min = y_min - 10
        # some small value
        y_max = y_max + 10

    # Determine Range
    y_range = y_max - y_min

    # Adjust ticks if needed
    if ticks < 2:
        ticks = 2
    else:
        if ticks > 2:
            ticks -= 2

    # Get raw step value
    temp_step = y_range / ticks

    # Calculate pretty step value
    mag = math.floor(np.log10(temp_step))
    mag_pow = math.pow(10, mag)
    mag_msd = int(temp_step / mag_pow + 0.5)
    step_size = mag_msd * mag_pow

    # build Y label array.
    # Lower and upper bounds calculations
    lb = step_size * math.floor(y_min / step_size)
    ub = step_size * math.ceil(y_max / step_size)

    # Build array
    val_ = lb
    while True:
        result.append(val_)
        val_ += step_size
        if val_ > ub:
            break

    return result


def _combine_files(report_files):
    df = None
    for filename in report_files:
        if df is None:
            df = _get_data_from_file(filename)
        else:
            df = pd.concat([df, _get_data_from_file(filename)])
    df.sort_index(inplace=True, axis=0)
    param_val = df[['TaskList[Parameter Estimation].(Problem)Parameter Estimation.Best Value']].idxmin().values[0]
    obj_val = df[['TaskList[Parameter Estimation].(Problem)Parameter Estimation.Best Value']].min().values[0]
    return df, obj_val, param_val


def _get_report_filemap(data_dir):
    """Returns the map of report files from the given data director

    :param data_dir: the directory to search for the update report files.
    :return: a map of id, and list of files for it
    :rtype: Dict[str, List[str]]
    """
    file_map = {}
    for filename in os.listdir(data_dir):
        if not (filename.endswith('_update_high.txt') or filename.endswith('_update_low.txt')):
            continue
        new_name = filename
        for replacement in [('__', '_'),
                            ('_update_high.txt', ''),
                            ('_update_low.txt', '')]:
            new_name = new_name.replace(replacement[0], replacement[1])
        pos = new_name.rfind('_')
        if pos < 0:
            continue

        index = new_name[pos + 1:]
        if index not in file_map:
            file_map[index] = []
        file_map[index].append(os.path.abspath(os.path.join(data_dir, filename)))
    return file_map


def _get_data_from_file(filename):
    """Returns the data from the given file als pandas dataframe

    :param filename: filename to report file
    :return: the pandas dataframe for the file
    :rtype: pd.DataFrame
    """
    df = pd.read_csv(filename, sep='\t', header=0)
    cols = df.columns.to_list()
    if 'Time' in cols:
        df = df.drop(labels=['Time'], axis=1, errors='ignore')
    df = df.set_index(df.columns[0])
    return df


_Problem = None
_Task = None
Arguments = {
    "lower_adjustment": '-50%',
    "upper_adjustment": '+50%',
    "modulation": 0.01,
    "tolerance": 1e-06,
    "iterations": 50,
    "scan_interval": 40,
    "disable_plots": True,
    "disable_tasks": True,
    "use_hooke": False,
    "prefix": "out_",
    "filename": ""
}
_DataModel = None


def _get_time_course():
    for index in range(_DataModel.getTaskList().size()):

        task = _DataModel.getTask(index)
        if isinstance(task, COPASI.CTrajectoryTask):
            return task

    return None


def _get_scan():
    for index in range(_DataModel.getTaskList().size()):

        task = _DataModel.getTask(index)
        if isinstance(task, COPASI.CScanTask):
            return task

    return None


def _get_fit_task():
    for index in range(_DataModel.getTaskList().size()):

        task = _DataModel.getTask(index)
        if isinstance(task, COPASI.CFitTask):
            return task

    return None


def _convert_opt_items(opt_items):
    opt_item_list = []
    for index in range(len(opt_items)):
        obj = _DataModel.getObject(COPASI.CCommonName(opt_items[index].getObjectCN()))
        opt_item_list.append(
            {
                'cn': opt_items[index].getObjectCN().getString(),
                'start_value': opt_items[index].getStartValue(),
                'lower_bound': opt_items[index].getLowerBound(),
                'upper_bound': opt_items[index].getUpperBound(),
                'index': index,
                'name': obj.getObjectDisplayName() if obj is not None else '',
            }
        )

    return opt_item_list


def _generate_scan_for_item(item, index, data_dir, update_model=False, lower=False, **kwargs):
    global _DataModel
    _DataModel.loadModel(Arguments["filename"])
    if Arguments["disable_plots"]:

        logger.info("... Disable other Plots")
        num = _DataModel.getPlotDefinitionList().size()
        for index1 in range(num):
            _DataModel.getPlotSpecification(index1).setActive(False)

    if Arguments["disable_tasks"]:
        logger.info("... Disable other Tasks")
        num = _DataModel.getTaskList().size()
        for index2 in range(num):
            _DataModel.getTask(index2).setScheduled(False)

    _get_time_course().getMethod().getParameter("Relative Tolerance").setDblValue(1E-09)
    fit_task = _get_fit_task()
    fit_task.setUpdateModel(update_model)

    if Arguments["use_hooke"]:
        fit_task.setMethodType(COPASI.CTaskEnum.Method_HookeJeeves)
        fit_task.getMethod().getParameter("Iteration Limit").setIntValue(Arguments["iterations"])
        fit_task.getMethod().getParameter("Tolerance").setDblValue(Arguments["tolerance"])
    else:
        fit_task.setMethodType(COPASI.CTaskEnum.Method_LevenbergMarquardt)
        fit_task.getMethod().getParameter("Iteration Limit").setIntValue(Arguments["iterations"])
        fit_task.getMethod().getParameter("Modulation").setDblValue(Arguments["modulation"])
        fit_task.getMethod().getParameter("Tolerance").setDblValue(Arguments["tolerance"])

    problem1 = fit_task.getProblem()
    problem1.setCalculateStatistics(False)
    logger.debug("... Remove OptItem")
    problem1.removeOptItem(item["index"])
    logger.debug("... Generate Scan")
    scan_task = _get_scan()
    scan_task.setScheduled(True)
    problem2 = scan_task.getProblem()
    problem2.setSubtask(COPASI.CTaskEnum.Task_parameterFitting)
    problem2.setContinueFromCurrentState(False)
    problem2.setOutputInSubtask(False)
    problem2.clearScanItems()
    problem2.addScanItem(1, Arguments["scan_interval"])
    scan_item = problem2.getScanItem(0)
    scan_item.getParameter("Object").setCNValue(COPASI.CRegisteredCommonName(item["cn"]))

    start_value = item["start_value"]
    param_sd = kwargs.get('param_sds', {}).get(item["name"], None)
    lower_value = _adjust_value(start_value, Arguments["lower_adjustment"],
                                item['lower_bound'], param_sd)
    upper_value = _adjust_value(start_value, Arguments["upper_adjustment"],
                                item['upper_bound'], param_sd)

    if not update_model:
        middle = "_noupdate"
        scan_item.getParameter("Maximum").setDblValue(upper_value)
        scan_item.getParameter("Minimum").setDblValue(lower_value)
    elif lower:
        middle = "_update_low"
        scan_item.getParameter("Minimum").setDblValue(start_value)
        scan_item.getParameter("Maximum").setDblValue(lower_value)
    else:
        middle = "_update_high"
        scan_item.getParameter("Maximum").setDblValue(upper_value)
        scan_item.getParameter("Minimum").setDblValue(start_value)

    scan_task.updateMatrices()
    logger.debug("... Generate Report")
    COPASI.COutputAssistant.createDefaultOutput(1251, scan_task, _DataModel)
    report_file = os.path.abspath("{0}/{1}_{2:05d}_{3}.txt".format(data_dir, Arguments['prefix'], index, middle))
    logger.debug(f"   Report target: '{report_file}'.")
    report = scan_task.getReport()
    report.setTarget(report_file)
    report.setAppend(False)
    report.setConfirmOverwrite(False)
    logger.debug("... Generate Plot")
    plot = COPASI.COutputAssistant.createDefaultOutput(251, scan_task, _DataModel)
    plot.setObjectName('{0} = {1}'.format('opt', start_value))
    out_file = os.path.abspath("{0}/{1}_{2:05d}_{3}.cps".format(data_dir, Arguments['prefix'], index, middle))
    logger.debug(f"Writing '{out_file}'.")
    _DataModel.saveModel(out_file, True)


def _adjust_value(value, adjustment, explicit=None, std_dev=None):
    """ adjusts the given value according to the multiplier

    :param value: the value to be adjusted

    :param adjustment: the adjustment to be made to the value
        adjustment is a float: the value will be adjusted to `value * adjustment`
        adjustment is a string starting with '+', the value will be adjusted to `value + value * adjustment`
        adjustment contains a '%', the adjustment value is adjusted by dividing by 100
        adjustment starts with '=', the value will be adjusted to `adjustment`
        adjustment is None, or 'default': the explicit value given will be used

    :param explicit: the explicit value to be used if adjustment is None or 'default'

    :return: the adjusted value
    """

    if not adjustment or adjustment == 'default':
        return float(explicit)

    if type(adjustment) is str and adjustment.endswith('SD') and std_dev is not None:
        return value + float(adjustment[:-2]) * std_dev

    if type(adjustment) is str:
        is_additive = adjustment.startswith('+')
        if is_additive:
            adjustment = adjustment[1:]

        is_declarative = adjustment.startswith('=')

        if is_declarative:
            adjustment = adjustment[1:]

        has_percent = '%' in adjustment
        if has_percent:
            adjustment = float(adjustment.rstrip('%')) / 100.0
            adjustment *= value

        adjustment = float(adjustment)
        if (adjustment < 0 or is_additive) and not is_declarative:
            return value + adjustment

        if has_percent or is_declarative:
            return adjustment

    return value * adjustment


def get_profiles_for_model(data_dir=None, **kwargs):
    """ Convenience function, combining the steps of preparing, running, and plotting

    :param data_dir: optional data directory to use, if not specified a temp directory will be used,
                     and deleted afterwards
    :param kwargs: optional parameters to use, all options from :func:`.prepare_files` and
                   :func:`.process_dir` can be specified.
    :return: the plot axis
    """
    delete_files = False

    if data_dir is None:
        delete_files = True
        data_dir = tempfile.mkdtemp()

    model = basico.get_current_model()
    old_filename = model.getFileName()
    filename = os.path.join(data_dir, 'model.cps')
    basico.save_model_and_data(filename)

    info = prepare_files(filename, data_dir, **kwargs)

    process_dir(data_dir,
                kwargs.get('pool_size', 4),
                kwargs.get('copasi_se', COPASI_SE),
                kwargs.get('max_time', None))

    result = plot_data(data_dir,
                       problem_size=(info['num_params'], info['num_data']),
                       **kwargs)
    if delete_files:
        shutil.rmtree(data_dir, ignore_errors=True)
    # restore old filename to ensure that relative files can still be found
    model.setFileName(old_filename)
    return result


def prepare_files(filename,
                  data_dir,
                  lower_adjustment='-50%',
                  upper_adjustment='+50%',
                  modulation=0.01,
                  tolerance=1e-06,
                  iterations=50,
                  scan_interval=40,
                  disable_plots=True,
                  disable_tasks=True,
                  use_hooke=False,
                  prefix="out_",
                  **kwargs):
    """Generates all files needed for the profile likelihood runs

    :param filename: filename of template copasi file. this file should already have a good fit, to be analyzed
    :param data_dir: directory in which the files will be generated
    :param lower_adjustment: adjustment of the current parameter value for the lower bound of the scan, this
           can be either a double multiplier, or a string with a percentage. By default, half of the current value
           will be substracted from it '-50%'. To take the current lower bound value as lower bound, use 'default'.
    :param upper_adjustment: adjustment of the current parameter value for the upper bound of the scan, this
           can be either a double multiplier, or a string with a percentage. By default, half of the current value
           will be added from it '+50%'. To take the current upper bound value as upper bound, use 'default'.
    :param modulation: In case Levenberg Marquardt is used, this is the modulation parameter, defaults to 0.01
    :param tolerance: optimization tolerance, defaults to 1e-6
    :param iterations: number of iterations to perform, defaults to 50
    :param scan_interval: number of scan intervals in each direction, defaults to 40, which means that 80 total
           optimization runs will be taken by default
    :param disable_plots: boolean, indicating whether other plots in the model should be disabled (defaults to True)
    :param disable_tasks: boolean, indicating whether other tasks in the model should be disabled (defaults to True)
    :param use_hooke: boolean indicating whether Hooks & Jeeves should be used (if True), or Levenberg Marquardt
           (if False) defaults to False.
    :param prefix: Prefix to be used for the files that will be generated in the `data_dir`, Defaults to 'out_'
    :return: dictionary, containing the number of parameters, the number of data sets, and the current objective value
    :rtype: dict
    """
    # remove old warnings
    COPASI.CCopasiMessage.clearDeque()

    global Arguments
    Arguments = locals()

    global _DataModel
    _DataModel = COPASI.CRootContainer.addDatamodel()
    if not _DataModel.loadModel(filename):
        logger.error("Could not load model")
        logger.error(COPASI.CCopasiMessage.getAllMessageText())
        sys.exit(2)

    logger.info(f"Copy Experimental Data to: {data_dir}")

    if os.path.abspath(filename) != os.path.abspath(os.path.join(data_dir, 'model.cps')):
        # copy files if needed
        basico.save_model_and_data(os.path.join(data_dir, 'original.cps'), model=_DataModel)
        _DataModel.loadModel(os.path.join(data_dir, 'original.cps'))

    # convert experiments to relative paths, so they'll work in cluster environment
    experiments = basico.save_experiments_to_dict(model=_DataModel)
    basico.load_experiments_from_dict(experiments, model=_DataModel)
    experiments_verify = basico.save_experiments_to_dict(model=_DataModel)

    global _Task
    _Task = _get_fit_task()
    if _Task is None:
        logger.error("No Fit Task.")
        return

    if not data_dir:
        data_dir = os.path.abspath(os.path.dirname(Arguments['prefix']))
    if not os.path.exists(data_dir):
        os.makedirs(data_dir, exist_ok=True)

    global _Problem
    _Problem = _Task.getProblem()

    num_params = _Problem.getOptItemSize()
    exp_set = _Problem.getExperimentSet()
    exp_set.compile(_DataModel.getModel().getMathContainer())
    num_data = exp_set.getDataPointCount()

    if 'obj' in kwargs:
        current_obj = kwargs['obj']
        param_sds = kwargs.get('param_sds', {})
    else:
        basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION, model=_DataModel)
        statistic = basico.get_fit_statistic(model=_DataModel, include_parameters=True)
        current_obj = statistic['obj']
        param_sds = {}
        for param in statistic['parameters']:
            param_sds[param['name']] = param['std_dev']

    info = {
        'num_params': num_params,
        'num_data': num_data,
        'obj': current_obj,
        'param_sds': param_sds,
    }

    with open(os.path.join(data_dir, '.info.yaml'), 'w', encoding='utf8') as f:
        yaml.dump(info, f)

    opt_item_list = _convert_opt_items(_Problem.getOptItemList())
    logger.debug(f"Have: {len(opt_item_list)} optItems")
    for index in range(len(opt_item_list)):
        logger.debug(f"Handling: {opt_item_list[index]['cn']}")
        # _generate_scan_for_item(optItemList[index], index, data_dir)
        _generate_scan_for_item(opt_item_list[index], index, data_dir, True, True, param_sds=param_sds)
        _generate_scan_for_item(opt_item_list[index], index, data_dir, True, param_sds=param_sds)

    # free the data model
    COPASI.CRootContainer.removeDatamodel(_DataModel)
    _DataModel = None
    return info


def printUsageAndExit():
    print("This program generates scans over each parameter of a parameter estimation. ")
    print()
    print("Usage: -f | --file <filename>")
    print("       -l | --lower <adjustment for current parameter value>")
    print("       -u | --upper <adjustment for current parameter value>")
    print("       -m | --modulation <levenberg marquardt modulation>")
    print("       -e | --tolerance <levenberg marquardt tolerance>")
    print("       -i | --iteration <iteration for parameter estimation run>")
    print("       -s | --steps <numer of steps for the scan>")
    print("       -o | --output-prefix <prefix for output files and reports>")
    print("       -p | --dont-disable-plots")
    print("       -t | --dont-disable-other-tasks")
    print("       -h | --help | /?")
    print()
    sys.exit(0)


def printStatus():
    print(f"Filename          : {Arguments['filename']}")
    print(f"Lower Adjustment  : {Arguments['lower_adjustment']}")
    print(f"Upper Adjustment  : {Arguments['upper_adjustment']}")
    print(f"modulation        : {Arguments['modulation']}")
    print(f"LM Toleratnce     : {Arguments['tolerance']}")
    print(f"iterations        : {Arguments['iterations']}")
    print(f"Disable Plots     : {Arguments['disable_plots']}")
    print(f"Disable Tasks     : {Arguments['disable_tasks']}")
    print(f"Output prefix     : {Arguments['prefix']}")
    print(f"Use Hooke & Jeeves: {Arguments['use_hooke']}")
    print()


def parse_args(args):
    global Arguments
    num_args = len(args)
    for index in range(num_args):
        lowerInvariant = args[index].strip().lower()
        s = args[index + 1] if index + 1 < num_args else None
        flag = s is not None
        if flag and (lowerInvariant == "-f" or lowerInvariant == "--file"):
            Arguments["filename"] = s
        elif flag and (lowerInvariant == "-o" or lowerInvariant == "--output-prefix"):
            Arguments["prefix"] = s
        elif flag and (lowerInvariant == "-l" or lowerInvariant == "--lower"):
            Arguments["lower_adjustment"] = float(s)
        elif flag and (lowerInvariant == "-u" or lowerInvariant == "--upper"):
            Arguments["upper_adjustment"] = float(s)
        elif flag and (lowerInvariant == "-m" or lowerInvariant == "--modulation"):
            Arguments["modulation"] = float(s)
        elif flag and (lowerInvariant == "-e" or lowerInvariant == "--tolerance"):
            Arguments["tolerance"] = float(s)
        elif flag and (lowerInvariant == "-i" or lowerInvariant == "--iteration"):
            Arguments["iterations"] = int(s)
        elif flag and (lowerInvariant == "-s" or lowerInvariant == "--steps"):
            Arguments["scan_interval"] = int(s)
        elif lowerInvariant == "-p" or lowerInvariant == "--dont-disable-plots":
            Arguments["disable_plots"] = False
        elif lowerInvariant == "-t" or lowerInvariant == "--dont-disable-other-tasks":
            Arguments["disable_tasks"] = False
        elif lowerInvariant == "-j" or lowerInvariant == "--use-hooke-and-jeeves":
            Arguments["use_hooke"] = True
        elif lowerInvariant == "/?" or lowerInvariant == "-h" or lowerInvariant == "--help":
            printUsageAndExit()


if __name__ == "__main__":
    print("generate_scans")
    print("==============")
    parse_args(sys.argv)
    if not os.path.exists(Arguments["filename"]):
        printUsageAndExit()

    printStatus()

    try:
        prepare_files(**Arguments)
    except:
        logger.exception("Exception occurred while preparing files")



import unittest
import basico
import basico.expressions

class TestExpressions(unittest.TestCase):

    def test_add(self):
        dm = basico.load_example('LM')
        basico.expressions.add_parameter_for('sum_rates', 'rates', 'sum')
        d = basico.as_dict(basico.get_parameters('sum_rates', exact=True))
        self.assertIsNotNone(d)
        self.assertEqual(d['expression'], 'S.Rate + E.Rate + ES.Rate + P.Rate')
        basico.expressions.add_parameter_for('sum_squares', 'rates', 'sum_of_squares')
        d = basico.as_dict(basico.get_parameters('sum_squares', exact=True))
        self.assertIsNotNone(d)
        self.assertEqual(d['expression'], 'S.Rate ^ 2 + E.Rate ^ 2 + ES.Rate ^ 2 + P.Rate ^ 2')
        basico.remove_datamodel(dm)

    def test_objects(self):
        dm = basico.load_example('LM')
        fluxes = basico.expressions._get_objects(dm, basico.expressions.OBJ.FLUXES)
        self.assertIsNotNone(fluxes)
        rates = basico.expressions._get_objects(dm, basico.expressions.OBJ.RATES_OF_CHANGE)
        self.assertIsNotNone(rates)
        concentrations = basico.expressions._get_objects(dm, basico.expressions.OBJ.CONCENTRATIONS)
        self.assertIsNotNone(concentrations)
        basico.remove_datamodel(dm)

    def test_min(self):
        e = basico.expressions._min([])
        self.assertEqual(e, '0')
        e = basico.expressions._min(['a'])
        self.assertEqual(e, '<a>')
        e = basico.expressions._min(['a', 'b'])
        self.assertEqual(e, 'min(<a>, <b>)')
        e = basico.expressions._min(['a', 'b', 'c'])
        self.assertEqual(e, 'min(<a>, min(<b>, <c>))')

    def test_max(self):
        e = basico.expressions._max([])
        self.assertEqual(e, '0')
        e = basico.expressions._max(['a'])
        self.assertEqual(e, '<a>')
        e = basico.expressions._max(['a', 'b'])
        self.assertEqual(e, 'max(<a>, <b>)')
        e = basico.expressions._max(['a', 'b', 'c'])
        self.assertEqual(e, 'max(<a>, max(<b>, <c>))')

    def test_sum(self):
        e = basico.expressions._sum([])
        self.assertEqual(e, '0')
        e = basico.expressions._sum(['a'])
        self.assertEqual(e, '<a>')
        e = basico.expressions._sum(['a', 'b'])
        self.assertEqual(e, '<a> + <b>')
        e = basico.expressions._sum(['a', 'b', 'c'])
        self.assertEqual(e, '<a> + <b> + <c>')

    def test_sum_of_squares(self):
        e = basico.expressions._sum_of_squares([])
        self.assertEqual(e, '0')
        e = basico.expressions._sum_of_squares(['a'])
        self.assertEqual(e, '<a>^2')
        e = basico.expressions._sum_of_squares(['a', 'b'])
        self.assertEqual(e, '<a>^2 + <b>^2')
        e = basico.expressions._sum_of_squares(['a', 'b', 'c'])
        self.assertEqual(e, '<a>^2 + <b>^2 + <c>^2')

    def test_sum_of_abs(self):
        e = basico.expressions._sum_abs([])
        self.assertEqual(e, '0')
        e = basico.expressions._sum_abs(['a'])
        self.assertEqual(e, 'abs(<a>)')
        e = basico.expressions._sum_abs(['a', 'b'])
        self.assertEqual(e, 'abs(<a>) + abs(<b>)')
        e = basico.expressions._sum_abs(['a', 'b', 'c'])
        self.assertEqual(e, 'abs(<a>) + abs(<b>) + abs(<c>)')

if __name__ == '__main__':
    unittest.main()



import unittest
import basico
import os

THIS_DIR = os.path.dirname(__file__)


class TestTimeCourse(unittest.TestCase):
    def setUp(self):
        basico.load_model(os.path.join(THIS_DIR, 'tellurium_sim.sbml'))

    def test_custom_output(self):
        tc = basico.run_time_course_with_output(['Time', '[S1]', '[S2]', 'Values[k1]'], 0, 5, 5)
        self.assertIsNotNone(tc)
        self.assertListEqual(tc.columns.to_list(), ['Time', '[S1]', '[S2]', 'Values[k1]'])

    def test_stochastic(self):
        basico.new_model(name='Simple Model')
        basico.add_reaction('R1', 'A ->')
        basico.add_reaction('R2', 'B -> B + C')
        basico.set_species('A', initial_particle_number=50)
        basico.set_species('B', initial_particle_number=10)
        basico.set_species('C', initial_particle_number=0)
        basico.set_reaction_parameters('(R1).k1', value=1)
        basico.set_reaction_parameters('(R2).k1', value=1)
        basico.add_event('E1', 'A <= 10', [('B', '50')])
        result = basico.run_time_course(duration=10, method='stochastic', use_numbers=True)
        self.assertIsNotNone(result)

        # test that we can set hybrid methods
        result = basico.run_time_course(duration=1, method='hybrid', use_numbers=True)
        self.assertIsNotNone(result)
        settings = basico.get_task_settings(basico.T.TIME_COURSE)
        self.assertEqual(settings['method']['name'], 'Hybrid (Runge-Kutta)')

    def test_stoch2(self):
        basico.new_model(name='Simple Model', substance_unit='1')
        basico.add_reaction('R1', 'A ->')
        basico.set_species('A', initial_particle_number=50)
        basico.set_reaction_parameters('(R1).k1', value=1)
        basico.add_event('E1', 'A.ParticleNumber <= 1', [('A.ParticleNumber', '50')])

        result = basico.run_time_course(duration=20, method='directMethod', use_numbers=True)
        self.assertIsNotNone(result)


if __name__ == '__main__':
    unittest.main()
